#!/usr/bin/env python

# stdlib imports
import io
import os.path

# third party imports
import numpy as np
import pytest
from esi_utils_rupture.factory import get_rupture
from esi_utils_rupture.origin import Origin
from openquake.hazardlib.gsim.chiou_youngs_2014 import ChiouYoungs2014

# local imports
from esi_shakelib.gmice.wgrw12 import WGRW12
from esi_shakelib.multigmpe import MultiGMPE
from esi_shakelib.station import StationList
from esi_shakelib.utils.exception import ShakeLibException
from esi_shakelib.utils.utils import get_extent, is_stable, replace_dyfi
from esi_shakelib.virtualipe import VirtualIPE

homedir = os.path.dirname(os.path.abspath(__file__))
datadir = os.path.join(homedir, "utils_data")


gmpe_cy14 = ChiouYoungs2014()
gmpe = MultiGMPE.__from_list__([gmpe_cy14], [1.0])
gmice = WGRW12()
ipe = VirtualIPE.__fromFuncs__(gmpe, gmice)


def test_replace_dyfi():
    stationfile = os.path.join(datadir, "nepal_dat.xml")
    dyfifile = os.path.join(datadir, "nepal_dyfi_dat.xml")

    original_stations = StationList.loadFromFiles([stationfile])
    dyfi_stations = StationList.loadFromFiles([dyfifile])
    dcursor = dyfi_stations.cursor
    dcursor.execute("SELECT count(*) from station")
    ndyfi1 = dcursor.fetchone()[0]

    original_cursor = original_stations.cursor
    original_cursor.execute("SELECT count(*) from station")

    original_cursor.execute(
        "SELECT count(*) from station WHERE instrumented=0"
    )
    noriginal_mmi = original_cursor.fetchone()[0]

    original_cursor.execute(
        "SELECT count(*) from station WHERE instrumented=0 and network='DYFI'"
    )
    noriginal_dyfi = original_cursor.fetchone()[0]

    noriginal_observed = noriginal_mmi - noriginal_dyfi

    stations = replace_dyfi(stationfile, dyfifile)

    scursor = stations.cursor
    scursor.execute(
        "SELECT count(*) from station where instrumented=0 and network != "
        "'DYFI'"
    )
    nobserved = scursor.fetchone()[0]

    assert nobserved == noriginal_observed
    scursor.execute(
        "SELECT count(*) from station where instrumented=0 and network='DYFI'"
    )
    ndyfi = scursor.fetchone()[0]
    assert ndyfi == ndyfi1


def test_get_extent_small_point():
    #
    # Do a point rupture
    # Small magnitude
    #
    config = {
        "extent": {"mmi": {"threshold": 4.0, "mindist": 100, "maxdist": 1000}}
    }
    eventfile = os.path.join(datadir, "event_wenchuan_small.xml")
    origin = Origin.fromFile(eventfile)
    rupture = get_rupture(origin)
    W, E, S, N = get_extent(config=config, ipe=ipe, rupture=rupture)
    np.testing.assert_allclose(W, 102.23333333333333)
    np.testing.assert_allclose(E, 104.53333333333333)
    np.testing.assert_allclose(S, 30.0)
    np.testing.assert_allclose(N, 31.966666666666667)


def test_get_extent_small_complex():
    #
    # Do a complex rupture
    #
    config = {
        "extent": {"mmi": {"threshold": 4.0, "mindist": 100, "maxdist": 1000}}
    }
    eventfile = os.path.join(datadir, "event_wenchuan.xml")
    origin = Origin.fromFile(eventfile)
    faultfile = os.path.join(datadir, "Hartzell11_fault.txt")
    rupture = get_rupture(origin, faultfile)
    W, E, S, N = get_extent(config=config, ipe=ipe, rupture=rupture)
    np.testing.assert_allclose(W, 96.383333333333333)
    np.testing.assert_allclose(E, 113.55)
    np.testing.assert_allclose(S, 24.25)
    np.testing.assert_allclose(N, 38.766666666666667)


def test_get_extent_bad_usage():
    #
    # Test bad usage
    #
    with pytest.raises(TypeError):
        W, E, S, N = get_extent()
    with pytest.raises(TypeError):
        W, E, S, N = get_extent(34)


def test_get_extent_aspect():
    #
    # Test ruptures with weird aspect ratios
    #
    config = {
        "extent": {"mmi": {"threshold": 4.0, "mindist": 100, "maxdist": 1000}}
    }
    eventfile = os.path.join(datadir, "event_wenchuan.xml")
    origin = Origin.fromFile(eventfile)
    #
    # Long horizontal rupture
    #
    rrep = io.StringIO(
        """
            100.0 30.0  0.0
            110.0 30.0 0.0
            110.0 30.0 5.0
            100.0 30.0 5.0
            100.0 30.0 0.0
            """
    )
    rupture = get_rupture(origin, rrep)
    W, E, S, N = get_extent(config=config, ipe=ipe, rupture=rupture)
    np.testing.assert_allclose(W, 93.733333333333333)
    np.testing.assert_allclose(E, 118.41666666666667)
    np.testing.assert_allclose(S, 20.833333333333333)
    np.testing.assert_allclose(N, 38.2)
    #
    # Long vertical rupture
    #
    rrep = io.StringIO(
        """
            100.0 24.0 0.0
            100.0 36.0 0.0
            100.0 36.0 5.0
            100.0 24.0 5.0
            100.0 24.0 0.0
            """
    )
    rupture = get_rupture(origin, rrep)
    W, E, S, N = get_extent(config=config, ipe=ipe, rupture=rupture)
    np.testing.assert_allclose(W, 89.366666666666667)
    np.testing.assert_allclose(E, 113.61666666666667)
    np.testing.assert_allclose(S, 17.283333333333333)
    np.testing.assert_allclose(N, 41.616666666666667)


def test_get_extent_stable_small():
    #
    # Do an event in a stable region
    # Small magnitude
    #
    config = {
        "extent": {"mmi": {"threshold": 4.0, "mindist": 100, "maxdist": 1000}}
    }
    eventfile = os.path.join(datadir, "event_oklahoma.xml")
    origin = Origin.fromFile(eventfile)
    rupture = get_rupture(origin)
    W, E, S, N = get_extent(config=config, ipe=ipe, rupture=rupture)
    np.testing.assert_allclose(W, -98.5)
    np.testing.assert_allclose(E, -96.28333333333333)
    np.testing.assert_allclose(S, 34.766666666666666)
    np.testing.assert_allclose(N, 36.583333333333336)


def test_get_extent_stable_large():
    #
    # Do an event in a stable region
    # Large magnitude
    #
    config = {
        "extent": {"mmi": {"threshold": 4.0, "mindist": 100, "maxdist": 1000}}
    }
    eventfile = os.path.join(datadir, "event_oklahoma_large.xml")
    origin = Origin.fromFile(eventfile)
    rupture = get_rupture(origin)
    W, E, S, N = get_extent(config=config, ipe=ipe, rupture=rupture)
    np.testing.assert_allclose(W, -104.45)
    np.testing.assert_allclose(E, -89.18333333333333)
    np.testing.assert_allclose(S, 29.3)
    np.testing.assert_allclose(N, 41.583333333333333)


# @pytest.mark.skip(reason="Function not currently used and needs strec configuration in the future")
def test_is_stable():
    assert is_stable(0, 0)  # Gulf of Guinea
    assert is_stable(-87.0, 36.0)  # Nashville
    assert not is_stable(-112.0, 39.0)  # Wasatch
    assert is_stable(135.0, -30.0)  # South Australia
    assert is_stable(180.0, 0.0)  # South Pacific (middle of ocean)
    assert is_stable(-60.0, -15.0)  # Brazil, near border of Bolivia


def test_extent_config():
    eventfile = os.path.join(datadir, "event_oklahoma_large.xml")
    origin = Origin.fromFile(eventfile)
    rupture = get_rupture(origin)

    config = {
        "extent": {
            "magnitude_spans": {"span1": [0, 6, 4, 3], "span2": [6, 10, 6, 4]}
        }
    }
    extent = get_extent(config=config, ipe=ipe, rupture=rupture)
    cmp_extent = (-99.4166667, -95.4083333, 32.675, 38.6833333)
    np.testing.assert_almost_equal(cmp_extent, extent)

    # Test for relative_offset
    config = {
        "extent": {
            "magnitude_spans": {"span1": [0, 6, 4, 3], "span2": [6, 10, 6, 4]},
            "relative_offset": [0.25, 0.5],
        }
    }
    extent = get_extent(config=config, ipe=ipe, rupture=rupture)
    cmp_extent = (-98.4166667, -94.4083333, 35.675, 41.6833333)
    np.testing.assert_almost_equal(cmp_extent, extent)

    config = {"extent": {"bounds": {"extent": [-100, 32, -95, 37]}}}
    extent = get_extent(config=config, ipe=ipe, rupture=rupture)
    cmp_extent = [-100, -95, 32, 37]
    np.testing.assert_almost_equal(extent, cmp_extent)


def test_exception():
    exception = ShakeLibException("Test exception")
    # Check __str__ override is correct
    assert str(exception) == "'Test exception'"


if __name__ == "__main__":
    test_replace_dyfi()
    test_get_extent_small_point()
    test_get_extent_small_complex()
    test_get_extent_bad_usage()
    test_get_extent_aspect()
    test_get_extent_stable_small()
    test_get_extent_stable_large()
    test_is_stable()
    test_extent_config()
    test_exception()



"""
Module to implement supported regional models from Kuehn et al. (2020).
This entire module can be removed (and its import into multigmpe.py) as
soon as the next version of openquake-engine hits pypi (v3.20.1 is the
current version).
"""

from openquake.hazardlib.gsim.base import add_alias
from openquake.hazardlib.gsim.kuehn_2020 import (
    KuehnEtAl2020SInter,
    KuehnEtAl2020SSlab,
    SUPPORTED_REGIONS,
    REGION_ALIASES,
)

for region in SUPPORTED_REGIONS[1:]:
    add_alias(
        "KuehnEtAl2020SInter" + REGION_ALIASES[region],
        KuehnEtAl2020SInter,
        region=region,
    )

for region in SUPPORTED_REGIONS[1:]:
    add_alias(
        "KuehnEtAl2020SSlab" + REGION_ALIASES[region],
        KuehnEtAl2020SSlab,
        region=region,
    )



#!/usr/bin/env python

# stdlib modules
import json
import os.path
import pathlib
import pickle
import sqlite3
import sys
import tempfile

# third party modules
import numpy as np
import pandas as pd

# local imports
from esi_shakelib.station import TABLES, StationList

homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?
shakedir = os.path.abspath(os.path.join(homedir, ".."))
sys.path.insert(0, shakedir)

#
# Set SAVE to True to write new versions of the output to disk,
# set it to False to actually run the tests.
#
SAVE = False


def test_station():
    homedir = os.path.dirname(os.path.abspath(__file__))

    #
    # First test the Calexico data on its own
    #
    event = "Calexico"

    datadir = os.path.abspath(os.path.join(homedir, "station_data"))
    datadir = os.path.abspath(os.path.join(datadir, event, "input"))

    inputfile = os.path.join(datadir, "stationlist_dat.xml")
    dyfifile = os.path.join(datadir, "ciim3_dat.xml")
    xmlfiles = [inputfile, dyfifile]

    stations = StationList.loadFromFiles(xmlfiles)

    df1, _ = stations.getStationDictionary(instrumented=True)
    df2, _ = stations.getStationDictionary(instrumented=False)

    ppath = os.path.abspath(os.path.join(datadir, "..", "database", "test1.pickle"))
    if SAVE:
        ldf = [df1, df2]
        with open(ppath, "wb") as f:
            pickle.dump(ldf, f, protocol=4)
    else:
        with open(ppath, "rb") as f:
            ldf = pickle.load(f)

        saved_df1 = ldf[0]
        saved_df2 = ldf[1]

        compare_dataframes(saved_df1, df1)
        compare_dataframes(saved_df2, df2)

    #
    # Should at least hit this code
    #
    imtlist = stations.getIMTtypes()
    assert "PGA" in imtlist
    assert "PGV" in imtlist

    #
    # Add the Northridge data to the Calexico data to test
    # addData()
    #
    event = "northridge"
    datadir = os.path.abspath(os.path.join(homedir, "station_data"))
    datadir = os.path.abspath(os.path.join(datadir, event, "input"))

    inputfile = os.path.join(datadir, "hist_dat.xml")
    dyfifile = os.path.join(datadir, "dyfi_dat.xml")
    len_stat1 = len(stations.getStationDictionary(instrumented=True)[0]["id"])
    xmlfiles = [inputfile, dyfifile]
    stations = stations.addData(xmlfiles, 3)
    len_stat2 = len(stations.getStationDictionary(instrumented=True)[0]["id"])

    # Check that more stations were added
    assert len_stat2 > len_stat1

    df1, _ = stations.getStationDictionary(instrumented=True)
    df2, _ = stations.getStationDictionary(instrumented=False)

    ppath = os.path.abspath(os.path.join(datadir, "..", "database", "test2.pickle"))
    if SAVE:
        ldf = [df1, df2]
        with open(ppath, "wb") as f:
            pickle.dump(ldf, f, protocol=4)
    else:
        with open(ppath, "rb") as f:
            ldf = pickle.load(f)

        saved_df1 = ldf[0]
        saved_df2 = ldf[1]

        compare_dataframes(saved_df1, df1)
        compare_dataframes(saved_df2, df2)


def test_station2():
    #
    # Test the wenchuan data
    #
    homedir = os.path.dirname(os.path.abspath(__file__))

    event = "wenchuan"

    datadir = os.path.abspath(os.path.join(homedir, "station_data"))
    datadir = os.path.abspath(os.path.join(datadir, event, "input"))

    inputfile = os.path.join(datadir, "stationlist.xml")
    xmlfiles = [inputfile]

    stations = StationList.loadFromFiles(xmlfiles)

    df1, _ = stations.getStationDictionary(instrumented=True)
    # Check Keys pressent
    assert "PGA" in _
    assert "PGV" in _
    assert "SA(0.3)" in _
    assert "SA(1.0)" in _
    assert "SA(3.0)" in _
    assert "PGV_sd" in df1
    assert "PGV" in df1
    assert "SA(0.3)" in df1
    assert "SA(1.0)" in df1
    assert "SA(3.0)" in df1
    assert "id" in df1
    df2, _ = stations.getStationDictionary(instrumented=False)
    # Check Keys pressent
    assert "MMI" in _
    ppath = os.path.abspath(os.path.join(datadir, "..", "database", "test3.pickle"))
    if SAVE:
        ldf = [df1, df2]
        with open(ppath, "wb") as f:
            pickle.dump(ldf, f, protocol=4)
    else:
        with open(ppath, "rb") as f:
            ldf = pickle.load(f)

        saved_df1 = ldf[0]
        saved_df2 = ldf[1]

        compare_dataframes(saved_df1, df1)
        compare_dataframes(saved_df2, df2)

        #
        # Dump the database to SQL and then restore it to a new
        # StationList object. Compare dataframes.
        #
        sql = stations.dumpToSQL()

        stations2 = StationList.loadFromSQL(sql)

        df1, _ = stations2.getStationDictionary(instrumented=True)
        df2, _ = stations2.getStationDictionary(instrumented=False)

        compare_dataframes(saved_df1, df1)
        compare_dataframes(saved_df2, df2)


def test_station3():
    #
    # Exercise the geojson code.
    #
    homedir = os.path.dirname(os.path.abspath(__file__))

    event = "wenchuan"

    datadir = os.path.abspath(os.path.join(homedir, "station_data"))
    datadir = os.path.abspath(os.path.join(datadir, event, "input"))

    inputfile = os.path.join(datadir, "stationlist.xml")
    xmlfiles = [inputfile]

    stations = StationList.loadFromFiles(xmlfiles)

    myjson = stations.getGeoJson()

    ofd = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
    jsonfile = ofd.name
    ofd.write(json.dumps(myjson).encode())
    ofd.close()

    stations2 = StationList.loadFromFiles([jsonfile])

    os.unlink(jsonfile)

    df1, _ = stations.getStationDictionary(instrumented=True)
    df2, _ = stations2.getStationDictionary(instrumented=True)
    compare_dataframes(df1, df2)

    df1, _ = stations.getStationDictionary(instrumented=False)
    df2, _ = stations2.getStationDictionary(instrumented=False)
    compare_dataframes(df1, df2)


def test_station4():
    homedir = os.path.dirname(os.path.abspath(__file__))

    event = "northridge"
    datadir = os.path.abspath(os.path.join(homedir, "station_data"))
    datadir = os.path.abspath(os.path.join(datadir, event, "input"))

    dyfifile = os.path.join(datadir, "dyfi_dat.xml")
    xmlfiles = [dyfifile]

    stations = StationList.loadFromFiles(xmlfiles)

    df1, _ = stations.getStationDictionary(instrumented=True)  # noqa
    df2, _ = stations.getStationDictionary(instrumented=False)  # noqa
    assert df1 is None


def test_station5():
    homedir = os.path.dirname(os.path.abspath(__file__))

    event = "Calexico"

    datadir = os.path.abspath(os.path.join(homedir, "station_data"))
    datadir = os.path.abspath(os.path.join(datadir, event, "input"))

    inputfile = os.path.join(datadir, "stationlist_dat.xml")
    dyfifile = os.path.join(datadir, "ciim3_dat.xml")

    xmlfiles = [inputfile, dyfifile]
    stations1 = StationList.loadFromFiles(xmlfiles)
    #
    # Load the data more than once to exercise the code that handles
    # repeated entries.
    #
    xmlfiles = [inputfile, inputfile, dyfifile, dyfifile]
    stations2 = StationList.loadFromFiles(xmlfiles)

    df1, _ = stations1.getStationDictionary(instrumented=True)
    df2, _ = stations2.getStationDictionary(instrumented=True)

    compare_dataframes(df1, df2)


def compare_dataframes(df1, df2):
    assert sorted(list(df1.keys())) == sorted(list(df2.keys()))

    idx1 = np.argsort(df1["id"])
    idx2 = np.argsort(df2["id"])

    for key in df1.keys():
        if df1[key].dtype in [np.float32, np.float64]:
            np.testing.assert_allclose(df1[key][idx1], df2[key][idx2], equal_nan=True)
        else:
            assert (df1[key][idx1] == df2[key][idx2]).all()


def test_packet():
    datadir = pathlib.Path(__file__).parent / "station_data" / "elmonte"
    db = sqlite3.connect(":memory:")
    stations_list = StationList(db)
    sta_set = set()
    imt_set = set()
    amp_set = set()
    min_nresp = 1
    shakemap_file = datadir / "ci38695658_groundmotions_dat.json"
    stations, amps, _ = stations_list._parse_shakemap_json(
        shakemap_file, min_nresp, sta_set, imt_set, amp_set
    )
    station_columns = [
        "station_id",
        "network",
        "code",
        "name",
        "lat",
        "lon",
        "elev",
        "vs30",
        "stddev",
        "instrumented",
        "source",
        "refid",
    ]
    amp_columns = [
        "station_id",
        "imt_type",
        "channel",
        "orientation",
        "amplitude",
        "stddev",
        "flag",
        "nresp",
    ]
    shakemap_stations_frame = pd.DataFrame(stations, columns=station_columns)
    shakemap_amps_frame = pd.DataFrame(amps, columns=amp_columns)

    sta_set = set()
    imt_set = set()
    amp_set = set()

    packet_file = datadir / "ci38695658_groundmotion_packet.json"
    stations, amps, _ = stations_list._parse_groundpacket_json(
        packet_file, sta_set, imt_set, amp_set
    )
    packet_stations_frame = pd.DataFrame(stations, columns=station_columns)
    packet_amps_frame = pd.DataFrame(amps, columns=amp_columns)
    for idx, shake_group in shakemap_amps_frame.groupby(["station_id", "imt_type"]):
        idx1 = packet_amps_frame["station_id"] == idx[0]
        idx2 = packet_amps_frame["imt_type"] == idx[1]
        idx3 = packet_amps_frame["channel"] != "ROTD50.0"
        packet_group = packet_amps_frame[(idx1) & (idx2) & (idx3)]
        shake_vals = np.sort(shake_group["amplitude"].values)
        packet_vals = np.sort(packet_group["amplitude"].values)
        np.testing.assert_allclose(shake_vals, packet_vals, rtol=1e-2)


if __name__ == "__main__":
    test_station()
    test_station2()
    test_station3()
    test_station4()
    test_station5()
    test_packet()



# Standard library imports
from abc import ABC, abstractmethod
import logging


class IMTConverter(ABC):
    """Base class for implementing conversions between intensity
    measurement types (IMT)."""

    @abstractmethod
    def convertAmps(self, imt_in, imt_out, imt):
        """
        Returns an array of converted IMT amplitude values.

        Args:
            imt_in (str): OQ intensity measure type. Same as type as the input
                values defined by the imt variable.
            imt_out (str): OQ intensity measure type that the values will
                be converted to.
            imt (OpenQuake IMT): The intensity measurements of the input
                ground motions. Valid IMTs are PGV, and SA.

        Returns:
            array: Numpy array of amps converted from imt_in to imt_out.

        Raises:
            ValueError: If not a valid conversion.
        """
        pass

    def getConversionFactor(self):
        """
        Helper method that returns the conversion factor.

        Returns:
            float: Conversion factor.
        """
        factor = self.conversion_factor
        return factor

    def getInputIMT(self, imt_out):
        """
        Get valid input IMT types that can be converted to the specified
        imt_out.

        Args:
            imt_out (str): OQ intensity measure type.

        Returns:
            list: List of valid input IMT types. If not available types
                None is returned.
        """
        imt_out = imt_out.upper().strip()
        try:
            imt_in = self.output_input[imt_out]
            return imt_in
        except KeyError:
            logging.info(f"No available conversion to {imt_out!r}")
            return None

    def _verifyConversion(self, imt_in, imt_out):
        """
        Helper method used to verify that a conversion is valid.

        Args:
            imt_in (str): OQ intensity measure type. Same as type as the input
                values defined by the imt variable.
            imt_out (str): OQ intensity measure type that the values will
                be converted to.

        Raises:
            ValueError: If the conversion is not valid.
        """
        valid_inputs = self.getInputIMT(imt_out)
        imt_in = imt_in.upper().strip()
        imt_out = imt_out.upper().strip()
        if imt_in not in valid_inputs and imt_in != imt_out:
            raise ValueError(f"No conversion available from {imt_in!r} to {imt_out!r}")



"""
Module to apply M9 and CB14 basin terms to modified GMPEs.
"""

import numpy as np

from openquake.hazardlib import const
from openquake.hazardlib.gsim.base import GMPE, registry
from openquake.hazardlib.gsim.mgmpe.cb14_basin_term import _get_cb14_basin_term

from esi_shakelib.utils.gmpe_coeffs import get_gmpe_coef_table


class M9BasinTerm(GMPE):
    """
    Implements the NSHM "M9" basin term to be applied to other GMMs for
    Cascadia.
    """

    REQUIRES_SITES_PARAMETERS = {"z2pt5"}
    REQUIRES_DISTANCES = set()
    REQUIRES_RUPTURE_PARAMETERS = set()
    DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = ""
    DEFINED_FOR_INTENSITY_MEASURE_TYPES = set()
    DEFINED_FOR_STANDARD_DEVIATION_TYPES = {const.StdDev.TOTAL}
    DEFINED_FOR_TECTONIC_REGION_TYPE = ""
    DEFINED_FOR_REFERENCE_VELOCITY = None

    m9 = False
    SJ = 0

    def __init__(self, gmpe_name, **kwargs):
        super().__init__(**kwargs)
        self.gmpe = registry[gmpe_name]()
        self.COEFFS = get_gmpe_coef_table(self.gmpe)
        self.set_parameters()

    def compute(self, ctx: np.recarray, imts, mean, sig, tau, phi):
        self.gmpe.compute(ctx, imts, mean, sig, tau, phi)
        for m, imt in enumerate(imts):
            fb = _get_cb14_basin_term(imt, ctx, self.SJ)
            if self.m9:
                fb_m9 = np.log(2.0)
                idx = ctx.z2pt5 > 6.0
                if imt.period > 1.9:
                    fb[idx] = fb_m9
            mean[m] += fb



"""
Module implements BooreKishida2017 class to convert between various
horizontal intensity measure components.
"""

# Standard imports
import glob
import logging
import os.path

import importlib.resources as importlib_resources

# Third party imports
import numpy as np
from openquake.hazardlib.const import IMC
from openquake.hazardlib.imt import PGA, PGV
import pandas as pd

# Local imports
from esi_shakelib.conversions.convert_imc import ComponentConverter


class BooreKishida2017(ComponentConverter):
    """
    This class implements the Boore and Kishida (2017) conversions for
    horizontal intensity measure components.

    This class explicitly supports the following subset of the conversions
    provided by B&K (specified as OpenQuake IMCs):

        - RotD50 <=> GMRotI50
        - RotD50 <=> GEOMETRIC_MEAN (i.e., Geometric Mean "as recorded")
        - RotD100 <=> RotD50
        - RotD50 <=> GREATER_OF_TWO_HORIZONTAL
        - RotD100 <=> GREATER_OF_TWO_HORIZONTAL
        - GMRotI50 <=> GREATER_OF_TWO_HORIZONTAL
        - GEOMETRIC_MEAN <=> GREATER_OF_TWO_HORIZONTAL

    Chain conversions are supported when using `convertAmps`. Otherwise
    conversions must be done in two+ steps using `convertAmpsOnce`. For IMCs
    not explicitly supported by B&K, we assume the IMC is equivalent
    to the geometric mean (which B&K call GM_AR).

    Notes
        - Assumes ALL unknown IMC types are GEOMETRIC_MEAN.

    References

        Boore, D.M. and T. Kishida (2017). Relations between some
        horizontal-component ground-motion intensity measures used
        in practice. Bulletin of the Seismological Society of
        America, 107(1), 334-343, doi: 10.1785/0120160250.

        Beyer, K., & Bommer, J. J. (2006). Relationships between median values
        and between aleatory variabilities for different definitions of the
        horizontal component of motion. Bulletin of the Seismological Society
        of America, 96(4A), 1512-1522.
    """

    def __init__(self, imc_in, imc_out):
        super().__init__()
        if isinstance(imc_in, IMC):
            self.imc_in = imc_in
        elif isinstance(imc_in, str):
            imc_from_str = self.imc_from_str(imc_in)
            if imc_from_str is None:
                logging.warning(
                    "Unknown IMC string '%s' in input, using Geometric Mean",
                    imc_in,
                )
                self.imc_in = IMC.GEOMETRIC_MEAN
            else:
                self.imc_in = imc_from_str
        else:
            logging.warning(
                "Unknown object passed as input %s, using Geometric Mean",
                type(imc_in),
            )
            self.imc_in = IMC.GEOMETRIC_MEAN
        if isinstance(imc_out, IMC):
            self.imc_out = imc_out
        elif isinstance(imc_out, str):
            imc_from_str = self.imc_from_str(imc_out)
            if imc_from_str is None:
                logging.warning(
                    "Unknown IMC string '%s' in output, using Geometric Mean",
                    imc_out,
                )
                self.imc_out = IMC.GEOMETRIC_MEAN
            else:
                self.imc_out = imc_from_str
        else:
            logging.warning(
                "Unknown object passed as opuput %s, using Geometric Mean",
                type(imc_out),
            )
            self.imc_out = IMC.GEOMETRIC_MEAN

        # Possible conversions
        self.conversion_graph = {
            IMC.RotD50: set(
                [
                    IMC.GMRotI50,
                    IMC.GEOMETRIC_MEAN,
                    IMC.RotD100,
                    IMC.GREATER_OF_TWO_HORIZONTAL,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                ]
            ),
            IMC.GMRotI50: set([IMC.RotD50, IMC.GREATER_OF_TWO_HORIZONTAL]),
            IMC.GEOMETRIC_MEAN: set(
                [IMC.RotD50, IMC.GREATER_OF_TWO_HORIZONTAL]
            ),
            IMC.RotD100: set([IMC.RotD50, IMC.GREATER_OF_TWO_HORIZONTAL]),
            IMC.GREATER_OF_TWO_HORIZONTAL: set(
                [
                    IMC.RotD50,
                    IMC.GMRotI50,
                    IMC.GEOMETRIC_MEAN,
                    IMC.RotD100,
                    IMC.RANDOM_HORIZONTAL,
                    IMC.HORIZONTAL,
                    IMC.MEDIAN_HORIZONTAL,
                ]
            ),
            IMC.HORIZONTAL: set([IMC.GREATER_OF_TWO_HORIZONTAL, IMC.RotD50]),
            IMC.MEDIAN_HORIZONTAL: set(
                [IMC.GREATER_OF_TWO_HORIZONTAL, IMC.RotD50]
            ),
            IMC.RANDOM_HORIZONTAL: set(
                [IMC.GREATER_OF_TWO_HORIZONTAL, IMC.RotD50]
            ),
        }
        # Check if any imc values are unknown. If they are, convert
        # to GEOMETRIC_MEAN
        self.checkUnknown()
        # Get shortest conversion "path" between imc_in and imc_out
        self.path = self.getShortestPath(
            self.conversion_graph, self.imc_in, self.imc_out
        )

    def convertAmpsOnce(self, imt, amps, rrups=None, mag=None):
        """
        Return an array of amps converted from one IMC to another.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.
            amps (array): A numpy array of the (logged) ground motions
                to be converted.
            rrups (array): A numpy array of the same shape as amps,
                containing the rupture distances of the ground motions.
            mag (float): The earthquake magnitude.

        Returns:
            array: A numpy array of converted ground motions (logged).

        Raises:
            ValueError: If mag and rrup are none or the IMT
            is not an allowed type.
        """
        # Check if mag and rrups are real values
        if mag is None or rrups is None:
            raise ValueError("No magnitude or rupture distances specified.")
        # Verify that the conversion is possible
        self._verifyConversion(self.imc_in, self.imc_out)
        # Return original amps if imc_in and imc_out are the same
        if self.pars is None:
            return amps.copy()
        # Get coeffecients
        (sigma, c0, r1, m1, m2) = self._getParamsFromIMT(imt)
        # Limit magnitude and rupture distances
        rrups_clipped = np.clip(rrups, 1e-2, 400)
        mag[mag < 2.0] = 2.0
        mag[mag > 9.0] = 9.0
        # Calculate conversion variable
        ln_ratio = (
            c0
            + r1 * np.log(rrups_clipped / 50)
            + m1 * (mag - 5.5)
            + m2 * (mag - 5.5) ** 2
        )
        #
        # The B&K file naming convention has things like D100D50, which
        # means the parameters give the (log) ratio of RotD100/RotD50,
        # but we use the convention that RotD100 would be the input IMC
        # and RotD50 would be the output IMC, so we reverse the sense
        # of the conversion here.
        #
        if self.forward:
            amps = amps - ln_ratio
        else:
            amps = amps + ln_ratio
        return amps

    def convertSigmasOnce(self, imt, sigmas):
        """
        Return an array of standard deviations converted from one IMC
        to another.

        Note that the action of this method is to always increase the
        input standard deviations. Thus, while converting from one IMC
        to another and then back again will yield the original ground
        motions via convertAmps(), the standard deviations will be
        inflated by both conversions via this method.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.
            sigmas (array): A numpy array of the standard deviations of
                the logged ground motions.

        Returns:
            array: A numpy array of converted standard deviations.

        Raises:
            ValueError: If mag and rrup are none or the IMT
            is not an allowed type.
        """
        # Verify that the conversion is possible
        self._verifyConversion(self.imc_in, self.imc_out)
        # Return original sigmas if imc_in and imc_out are the same
        if self.pars is None:
            return sigmas.copy()
        # Get coeffecients
        (sigma, c0, r1, m1, m2) = self._getParamsFromIMT(imt)
        # Calculate conversion
        sigmas = np.sqrt(sigmas**2 + sigma**2)

        return sigmas

    def _getParamsFromIMT(self, imt):
        """
        Helper function to return (possibly interpolated) conversion
        parameters for a given IMT.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.

        Returns:
            (float, float, float, float, float): Coeffients for conversion.
        """
        if imt == PGA():
            sigma = self.pars["sigma"][0]
            c0 = self.pars["c0smooth"][0]
            r1 = self.pars["r1smooth"][0]
            m1 = self.pars["m1smooth"][0]
            m2 = self.pars["m2smooth"][0]
        elif imt == PGV():
            sigma = self.pars["sigma"][1]
            c0 = self.pars["c0smooth"][1]
            r1 = self.pars["r1smooth"][1]
            m1 = self.pars["m1smooth"][1]
            m2 = self.pars["m2smooth"][1]
        elif "SA" in imt.string:
            imt_per = imt.period
            pa = self.pars["per"][2:]
            sigma = np.interp(imt_per, pa, self.pars["sigma"][2:])
            c0 = np.interp(imt_per, pa, self.pars["c0smooth"][2:])
            r1 = np.interp(imt_per, pa, self.pars["r1smooth"][2:])
            m1 = np.interp(imt_per, pa, self.pars["m1smooth"][2:])
            m2 = np.interp(imt_per, pa, self.pars["m2smooth"][2:])
        else:
            raise ValueError(f"Unknown IMT: {str(imt)}")
        return (sigma, c0, r1, m1, m2)

    @staticmethod
    def _imcPairToFile(imc_in, imc_out):
        """
        Helper function to find the name of the file representing
        the conversion.

        Returns:
            (str, bool): The filename and a boolean 'forward' indicating
            whether the conversion should be done in the forward (True)
            or inverse (False) direction. If filename is None, then no
            appropriate conversion file coule be found; if it is the
            string 'Null', then imc_in and imc_out evaluated to be the
            same.
        """
        datadir = (
            importlib_resources.files("esi_shakelib.conversions.imc") / "data"
        )
        conv_files = glob.glob(os.path.join(datadir, "*.csv"))
        stub1 = BooreKishida2017._imcToFilestr(imc_in)
        stub2 = BooreKishida2017._imcToFilestr(imc_out)
        if stub1 == stub2:
            # No conversion necessary
            return ("Null", True)
        #
        # Look for the conversion from imc_in -> imc_out
        #
        stub = stub1 + stub2
        filelist = [name for name in conv_files if stub in name]
        if len(filelist) == 1:
            return (filelist[0], True)
        #
        # Now try the conversion from imc_out -> imc_in
        #
        stub = stub2 + stub1
        filelist = [name for name in conv_files if stub in name]
        if len(filelist) == 1:
            return (filelist[0], False)
        #
        # Can't find anything
        #
        return (None, None)

    @staticmethod
    def _imcToFilestr(oq_imc):
        """
        Helper function to convert an OpenQuake IMC into part of the
        Boore & Kishida file name.
        """

        if oq_imc == IMC.RotD50:
            return "D50"
        elif oq_imc == IMC.RotD100:
            return "D100"
        elif oq_imc == IMC.GMRotI50:
            return "GM50"
        elif (
            oq_imc == IMC.GEOMETRIC_MEAN
            or oq_imc == IMC.HORIZONTAL
            or oq_imc == IMC.RANDOM_HORIZONTAL
            or oq_imc == IMC.MEDIAN_HORIZONTAL
        ):
            return "GMAR"
        elif oq_imc == IMC.GREATER_OF_TWO_HORIZONTAL:
            return "Larger"
        else:
            #
            # For less common IMCs, Beyer & Bommer (2006) found most
            # of them to be more or less equivalent to geometric mean
            #
            logging.warning(f"Can't handle IMC {oq_imc}, using GMAR")
            return "GMAR"

    def _verifyConversion(self, imc_in, imc_out=None):
        """
        Helper method to ensure that the conversion is possible.

        Args:
            imc_in (IMC): OpenQuake IMC type of the input amp array.
            imc_out (IMC): Desired OpenQuake IMC type of the output amps.
                Default is None.

        Raises:
            ValueError if imc_in or imc_out are not valid..
        """
        filename, forward = self._imcPairToFile(imc_in, imc_out)
        if filename is None:
            raise ValueError(
                f"Can't find a conversion file for {imc_in} and {imc_out}"
            )
        self.forward = forward
        if filename == "Null":
            # Null conversion -- imc_in and imc_out are either identical
            # or at least functionally equivalent
            self.pars = None
        else:
            self.pars = pd.read_csv(filename)



import copy

import numpy as np

from openquake.hazardlib.gsim.base import GMPE
from openquake.hazardlib.imt import PGA, PGV, SA, MMI
from openquake.hazardlib import const

from esi_shakelib.utils.exception import ShakeLibException
from esi_shakelib.generic_site_amplitication import GenericSiteAmplification
from esi_utils_rupture.point_rupture import PointRupture


class VirtualIPE(GMPE):
    """
    Implements a virtual IPE that is the combination of a MultiGMPE
    and a GMICE. Will first attempt to use PGV for the intensities,
    and then will try PGA, and then SA(1.0), and then will bail out.

    Uncertainty is computed by combining the uncertainty of the GMPE
    with the uncertainty of the GMICE. Standard error propagation
    techniques are used (see the ShakeMap manual for a detailed
    explanation). For the intra- and inter-event components of total
    uncertainty, we assign all of GMICE uncertaninty to the intra-event
    term, and none to the inter-event term. This choice is conservative,
    and seems appropriate until GMICE are produced with separate inter-
    and intra-event terms.

    Note that the combined inter- and intra-event uncertainties will
    only approximately equal the total uncertainty because the GMPEs
    will only produce combined uncertainties that are approximately
    equal to their total uncertainty.

    """

    #: The OpenQuake IMT this module can produce ('MMI' only).
    DEFINED_FOR_INTENSITY_MEASURE_TYPES = set([MMI])
    #: The OpenQuake standard deviation types that may be produced (will
    #: depend on the GMPE provided).
    DEFINED_FOR_STANDARD_DEVIATION_TYPES = set([const.StdDev.TOTAL])
    #: Distance measures required (will depend on the GMPE provided).
    REQUIRES_DISTANCES = None
    #: OpenQuake IMC used (will depend on the GMPE, but "Larger" is
    #: typical).
    DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = None
    #: Determined by the GMPE selected.
    DEFINED_FOR_TECTONIC_REGION_TYPE = None
    #: Determined by the GMPE selected.
    REQUIRES_RUPTURE_PARAMETERS = None
    #: Determined by the GMPE selected.
    REQUIRES_SITES_PARAMETERS = None

    @classmethod
    def __fromFuncs__(cls, gmpe, gmice, rupture_obj=None):
        """
        Creates a new VirtualIPE object with the specified MultiGMPE and
        GMICE. There is no default constructor, you must use this method.

        Args:
            gmpe: An instance of the MultiGMPE object.
            gmice: An instance of a GMICE object.
            rupture_obj: A rupture object.

        Returns:
            :class:`VirtualIPE`: A new instance of a VirtualIPE object.

        """
        self = cls()
        self.gmpe = gmpe
        self.gmice = gmice
        self.rupture_obj = rupture_obj

        if gmpe.ALL_GMPES_HAVE_PGV is True and any(
            "PGV" in t.__name__ for t in gmice.DEFINED_FOR_INTENSITY_MEASURE_TYPES
        ):
            self.imt = PGV()
        elif any(
            "PGA" in t.__name__ for t in gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES
        ) and any(
            "PGA" in t.__name__ for t in gmice.DEFINED_FOR_INTENSITY_MEASURE_TYPES
        ):
            self.imt = PGA()
        elif any(
            "SA" in t.__name__ for t in gmpe.DEFINED_FOR_INTENSITY_MEASURE_TYPES
        ) and any(
            "SA" in t.__name__ for t in gmice.DEFINED_FOR_INTENSITY_MEASURE_TYPES
        ):
            self.imt = SA(1.0)
        else:
            raise ShakeLibException(
                "The supplied GMPE and GMICE do not have a common IMT"
            )

        self.DEFINED_FOR_STANDARD_DEVIATION_TYPES = (
            gmpe.DEFINED_FOR_STANDARD_DEVIATION_TYPES.copy()
        )

        self.REQUIRES_DISTANCES = gmpe.REQUIRES_DISTANCES.copy()
        self.REQUIRES_RUPTURE_PARAMETERS = gmpe.REQUIRES_RUPTURE_PARAMETERS.copy()
        self.REQUIRES_SITES_PARAMETERS = gmpe.REQUIRES_SITES_PARAMETERS.copy()
        self.DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = copy.copy(
            gmpe.DEFINED_FOR_INTENSITY_MEASURE_COMPONENT
        )
        self.DEFINED_FOR_TECTONIC_REGION_TYPE = copy.copy(
            gmpe.DEFINED_FOR_TECTONIC_REGION_TYPE
        )

        return self

    def get_mean_and_stddevs(self, sx, rx, dx, imt, stddev_types, fd=None):
        """
        See superclass
        `method <http://docs.openquake.org/oq-hazardlib/master/gsim/index.html#openquake.hazardlib.gsim.base.GroundShakingIntensityModel.get_mean_and_stddevs>`__
        for parameter definitions. The only acceptable IMT is MMI.

        Additional subclass argument is "fd", which is the directivity
        amplification factor in natural log units. This is optional,
        and must be a numpy array with the same dimentions as the
        sites and is added to the ground motions before conversion to
        MMI.

        Returns:
            ndarray, list of ndarray:

            mmi (ndarray): Ground motions predicted by the MultiGMPE using
            the supplied parameters are converted to MMI using the GMICE.

            mmi_sd (list of ndarrays): The uncertainty of the combined
            prediction/conversion process. The prediction uncertainty will
            typically be either OpenQuake's TOTAL or INTRA_EVENT.  But can
            be any set that the MultiGMPE supports. See the ShakeMap manual
            for a detailed discussion of the way the uncertainty is computed.

        """  # noqa

        if isinstance(imt, list):
            imt = imt[0]
        if imt != MMI():
            raise ValueError("imt must be MMI")
        #
        # Get the mean ground motions and stddev for the preferred IMT
        #
        mgm, sdev = self.gmpe.get_mean_and_stddevs(sx, rx, dx, self.imt, stddev_types)

        if fd is not None:
            mgm = mgm + fd

        #
        # Get the MMI and the dMMI/dPGM from the GMICE
        #
        if hasattr(dx, "rrup"):
            dist4gmice = dx.rrup
        else:
            dist4gmice = dx.rhypo

        mmi, dmda = self.gmice.getMIfromGM(mgm, self.imt, dist4gmice, rx.mag)

        #
        # Compute the uncertainty of the combined prediction/conversion
        # Total and intra-event uncertanty are inflated by the uncertainty
        # of the conversion; inter-event uncertainty is not.
        #
        nsd = len(sdev)
        mmi_sd = [None] * nsd
        gm2mi_var = (self.gmice.getGM2MIsd()[self.imt]) ** 2
        dmda *= dmda
        for i in range(nsd):
            gm_var_in_mmi = dmda[0] * sdev[i] ** 2
            if stddev_types[i] == const.StdDev.INTER_EVENT:
                mmi_sd[i] = np.sqrt(gm_var_in_mmi)
            else:
                mmi_sd[i] = np.sqrt(gm2mi_var + gm_var_in_mmi)
        return mmi, mmi_sd



#!/usr/bin/env python

# Standard library imports
import os.path
import sys

# Third party imports
import pytest

# Local imports
from esi_shakelib.conversions.imt.bommer_alarcon_2006 import BommerAlarcon2006


homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?
shakedir = os.path.abspath(os.path.join(homedir, "..", "..", ".."))
sys.path.insert(0, shakedir)


def test_bommeralarcon2006():
    # Inputs
    PGVin = 10
    PSA05in = 0.1
    ba06 = BommerAlarcon2006()

    # Test that the correct inputs are returned for valid outputs
    input1 = ba06.getInputIMT("pgv  ")
    input2 = ba06.getInputIMT("PGV")
    input3 = ba06.getInputIMT("psa05  ")
    input4 = ba06.getInputIMT("PSA05")
    assert len(input1) == len(input2) == 1
    assert input1[0] == input2[0] == "PSA05"
    assert len(input3) == len(input4) == 1
    assert input3[0] == input4[0] == "PGV"

    # Test that invalid outputs return None
    input5 = ba06.getInputIMT("INVALID")
    assert input5 == None

    # Test valid conversions
    PSA05out = ba06.convertAmps("pgv", "PSA05", PGVin)
    PGVout = ba06.convertAmps("PSA05", "PGV", PSA05in)
    vfact = ba06.getConversionFactor()
    assert abs(PSA05out - 0.2038735983690112) < 0.0001
    assert abs(PGVout - 4.905000) < 0.001
    assert abs(vfact - 49.050000) < 0.001

    # Test invalid conversions types
    with pytest.raises(Exception) as a:
        tmp = ba06.convertAmps("INVALID", "PSA05", PGVin)
    with pytest.raises(Exception) as a:
        tmp = ba06.convertAmps("PGV", "INVALID", PGVin)
    with pytest.raises(Exception) as a:
        tmp = ba06.convertAmps("INVALID", "INVALID", PGVin)


if __name__ == "__main__":
    test_bommeralarcon2006()



from abc import ABC, abstractmethod


class CrossCorrelationBase(ABC):
    """
    The abstract base class of cross-correlation functions.
    """

    @abstractmethod
    def __init__(self, periods):
        """The derived class should take an array of periods as its single
        argument. These are the periods to which the index arrays of the
        getCorrelation() method will refer.
        """
        pass  # pragma: no cover

    @abstractmethod
    def getCorrelation(self, ix1, ix2, h):
        """
        Compute the correlation between two periods and a separation distance
        of h km. ix1 and ix2 give the indices into the array of periods that
        is provided to the class constructor. The code is implemented in this
        somewhat awkward way for performance reasons. See the loth_baker_2013
        module for an example.

        The index arrays (ix1 and ix2) and h array must have the same
        dimensions. The
        indices may be equal, and there is no restriction on which one is
        larger. The indices refer to periods in the 'period' argument to the
        class constructor.

        Args:
            ix1 (ndarray):
                The indices of the first period of interest.
            ix2 (ndarrays):
                The indices of the second period of interest.
            h (ndarray):
                The separation distance between two sites (units of km).
                The result is placed in h, so h must be copied if it is
                to be preserved.

        Returns:
            h (ndarray): The predicted correlation coefficient. The output
            array will have the same shape as the inputs.

        """
        pass  # pragma: no cover



#!/usr/bin/env python

from esi_shakelib.utils.imt_string import oq_to_file, file_to_oq


def test_imt():
    """Test the imt string functions."""
    assert oq_to_file("SA(1.0)") == "psa1p0"
    assert oq_to_file("SA(0.3)") == "psa0p3"
    assert oq_to_file("SA(15.0)") == "psa15p0"
    assert oq_to_file("SA(3)") == "psa3p0"
    assert oq_to_file("SA(.5)") == "psa0p5"

    try:
        _ = oq_to_file("SA()")
    except ValueError as ve:
        assert 1 == 1

    assert file_to_oq("psa1p0") == "SA(1.0)"
    assert file_to_oq("psa0p3") == "SA(0.3)"
    assert file_to_oq("psa15p0") == "SA(15.0)"

    try:
        _ = file_to_oq("psa2")
    except ValueError as ve:
        assert 1 == 1

    try:
        _ = file_to_oq("psa2p")
    except ValueError as ve:
        assert 1 == 1

    # Test that a fileimt that is the same as ['PGA', 'PGV', 'MMI']
    # is simply returned
    assert file_to_oq("pga") == "PGA"
    assert file_to_oq("pgv") == "PGV"
    assert file_to_oq("mmi") == "MMI"


if __name__ == "__main__":
    test_imt()



# stdlib imports
import copy
import json
import logging
import re
import sqlite3
import xml.etree.ElementTree as ET
from collections import OrderedDict

# third party imports
import numpy as np
import pandas as pd
from gmpacket.packet import GroundMotionPacket
from scipy import constants

# local imports


TABLES = OrderedDict(
    (
        (
            "station",
            OrderedDict(
                (
                    ("id", "text primary key"),  # id is net.sta
                    ("refid", "int"),  # foreign "key" to reference table
                    ("network", "text"),
                    ("code", "text"),
                    ("name", "text"),
                    ("lat", "float"),
                    ("lon", "float"),
                    ("elev", "float"),
                    ("vs30", "float"),
                    ("stddev", "float"),
                    ("instrumented", "int"),
                    ("source", "text"),
                )
            ),
        ),
        (
            "imt",
            OrderedDict((("id", "integer primary key"), ("imt_type", "text"))),
        ),
        (
            "amp",
            OrderedDict(
                (
                    ("id", "integer primary key"),
                    ("station_id", "text"),
                    ("imt_id", "int"),
                    ("original_channel", "text"),
                    ("orientation", "text"),
                    ("amp", "float"),
                    ("stddev", "float"),
                    ("nresp", "int"),
                    ("flag", "text"),
                )
            ),
        ),
        (
            "reference",
            OrderedDict(
                (
                    ("id", "integer primary key"),
                    ("shortref", "text"),
                    ("longref", "text"),
                    ("description", "text"),
                )
            ),
        ),
    )
)

#
# These are the netid's that indicate MMI data
#
CIIM_TUPLE = ("dyfi", "mmi", "intensity", "ciim")
NON_CHANNEL_PATTERNS = ["GREATER"]
SUPPORTED_METRICS = ["PGA", "PGV", "SA"]
ORIENTATIONS = {"1": "h", "2": "h", "Z": "v"}
SUPPORTED_SA_PERIODS = [0.3, 1.0, 3.0]

PSA_TO_SA = {
    "psa03": "SA(0.3)",
    "psa10": "SA(1.0)",
    "psa30": "SA(3.0)",
}


def convert_units(imt_type, input_units, amplitude):
    """
    Convert amp units based on units strings.
    """
    if re.search(r"PGA|SA", imt_type) is not None:
        # convert to ln(g)7
        if input_units in ["cm/s^2", "cm/s**2", "cm/s/s", "gals"]:
            amplitude = np.log(amplitude / constants.g / 100)
        elif input_units in ["m/s^2", "m/s**2", "m/s/s"]:
            amplitude = np.log(amplitude / constants.g)
        elif input_units == "g":
            amplitude = np.log(amplitude)
        elif input_units == "%g":
            amplitude = np.log(amplitude / 100)
        elif input_units in ["ln(g)"]:
            pass
        else:
            raise ValueError(f"Unknown acceleration input units {input_units}")
    elif imt_type == "PGV":
        if input_units in ["cm/s", "cms"]:
            amplitude = np.log(amplitude)
        elif input_units in ["m/s"]:
            amplitude = np.log(amplitude * 100)
        elif input_units in ["ln(cm/s)"]:
            pass
        else:
            raise ValueError(f"Unknown velocity input units {input_units}")
    elif re.search(r"FAS", imt_type) is not None:
        if input_units in ["cm/s", "cms"]:
            amplitude = np.log(amplitude)
        elif input_units in ["m/s"]:
            amplitude = np.log(amplitude * 100)
        elif input_units == "ln(cm/s)":
            pass
        else:
            raise ValueError(f"Unknown FAS input units {input_units}")
    elif re.search(r"DURATION", imt_type) is not None:
        if input_units in ["s", "sec", "seconds"]:
            amplitude = np.log(amplitude)
        elif input_units in ["ln(s)", "ln(sec)", "ln(seconds)"]:
            pass
        else:
            raise ValueError(f"Unknown DURATION input units {input_units}")
    elif imt_type == "ARIAS":
        if input_units in ["cm/s", "cms"]:
            amplitude = np.log(amplitude / 100)
        elif input_units in ["m/s"]:
            amplitude = np.log(amplitude)
        elif input_units in ["ln(m/s)"]:
            pass
        else:
            raise ValueError(f"Unknown velocity input units {input_units}")
    else:
        raise ValueError(f"Unknown imt_type {imt_type}")
    return amplitude


class StationList(object):
    """
    A class to facilitate reading ShakeMap formatted XML fies of peak
    amplitudes and MMI, and
    produce tables of station data. Seismic stations are considered to
    be 'instrumented'; MMI data is not instrumented and is indicated
    in the ShakeMap XML with a ``netid`` attribute of "DYFI," "MMI,"
    "INTENSITY," or "CIIM."

    .. note::
      Typically the user will call the class method :meth:`fromXML`
      to create a :class:`StationList` object the first time
      a set of station files are processed. (Or, as an alternative,
      the user can call :meth:`loadFromXML` and :meth:`fillTables`
      sequentially.)
      This will create a database at the location specified by the
      ``dbfile`` parameter to :meth:`fromXML`. Subsequent programs
      can use the default constructor to simply load ``dbfile``.

    """

    def __init__(self, db):
        """
        The default constructor reads a pre-built SQLite database of
        station data.

        Args:
            dbfile (str):
                A SQLite database file containing pre-processed
                station data.

        Returns:
            A :class:`StationList` object.

        """
        self.db = db
        self.cursor = self.db.cursor()

    def __del__(self):
        """
        Closes out the database when the object is destroyed.
        """
        self.db.commit()
        self.cursor.close()
        self.db.close()

    @classmethod
    def loadFromSQL(cls, sql, dbfile=":memory:"):
        """
        Create a new object from saved SQL code (see :meth:`dumpToSQL`).

        Args:
            sql (str):
                SQL code to create and populate the database
            dbfile (str):
                The path to a file in which the database will reside.
                The default is ':memory:' for an in-memory database.

        Returns:
            :class:`Stationlist` object.
        """
        db = sqlite3.connect(dbfile, timeout=15)
        self = cls(db)
        self.cursor.executescript(sql)
        return self

    def dumpToSQL(self):
        """
        Dump the database as a string of SQL code (see :meth:`loadFromSQL`).

        Args:
            None

        Returns:
            A string of SQL sufficient to restore and repopulate the
            database.
        """

        return "\n".join(list(self.db.iterdump()))

    @classmethod
    def loadFromFiles(cls, filelist, min_nresp=3, dbfile=":memory:"):
        """
        Create a StationList object by reading one or more ShakeMap XML or
        JSON input files.

        Args:
            filelist (sequence of str):
                Sequence of ShakeMap XML and/or JSON input files to read.
            min_nresp (int):
                The minimum number of DYFI observations required to form and
                valid observation. Default is 3.
            dbfile (str):
                Path to a file into which to write the SQLite database.
                The default is ':memory:' for an in-memory database.

        Returns:
            :class:`StationList` object
        """
        # Create the database and tables
        db = sqlite3.connect(dbfile, timeout=15)
        self = cls(db)
        self._createTables()
        self.addData(filelist, min_nresp)
        return self

    def addData(self, filelist, min_nresp):
        """
        Add data from XML or JSON files to the existing StationList.

        Args:
            filelist:
                A list of ShakeMap XML or JSON input files.
            min_nresp (int):
                The minimum number of DYFI observations required to form and
                valid observation.

        Returns:
            nothing: Nothing.
        """
        jsonfiles = [x for x in filelist if x.endswith(".json")]
        xmlfiles = [x for x in filelist if x.endswith(".xml")]
        if len(jsonfiles):
            self._loadFromJSON(jsonfiles, min_nresp)
        if len(xmlfiles):
            self._loadFromXML(xmlfiles, min_nresp)
        self._fixOrientations()
        return self

    def _loadFromXML(self, xmlfiles, min_nresp):
        """
        Create a StationList object by reading one or more ShakeMap XML input
        files.

        Args:
            xmlfiles (sequence of str):
                Sequence of ShakeMap XML input files to read.
            min_nresp (int):
                The minimum number of DYFI responses for an observation to be
                included in the station output.

        Returns:
            nothing: Nothing.
        """
        # Parse the xml into a dictionary
        stationdict = {}
        imtset = set()
        for xmlfile in xmlfiles:
            stationdict, ims = self._filter_station(
                xmlfile, stationdict, min_nresp
            )
            imtset |= ims
        # fill the database and create the object from it
        self._loadFromDict(stationdict, imtset)
        return

    def _parse_shakemap_json(
        self, jfile, min_nresp, sta_set, imt_set, amp_set
    ):
        with open(jfile, "r", encoding="utf-8") as jfp:
            stas = json.load(jfp)

        if "type" not in stas:
            logging.warning(
                "%s appears to contain no stations, skipping", jfile
            )
            return ([], [], [])
        if stas["type"] != "FeatureCollection":
            logging.warning(
                "%s is not a ShakeMap JSON stationlist, skipping", jfile
            )
            return ([], [], [])

        # if present, insert reference stuff into the database
        reference_rows = []
        if "references" in stas:
            for shortref, refdict in stas["references"].items():
                longref = refdict["long_reference"]
                description = refdict["description"]
                reference_rows.append([shortref, longref, description])

        station_rows = []
        amp_rows = []
        for feature in stas["features"]:
            sta_id = feature["id"]
            if sta_id in sta_set:
                continue
            else:
                sta_set.add(sta_id)
            lon = feature["geometry"]["coordinates"][0]
            lat = feature["geometry"]["coordinates"][1]
            netid = feature["properties"]["network"]
            code = sta_id.replace(netid + ".", "")
            network = feature["properties"]["network"]
            name = feature["properties"].get("name", None)
            elev = feature["properties"].get("elev", None)
            vs30 = feature["properties"].get("vs30", None)
            source = feature["properties"].get("source", None)
            refid = feature["properties"].get("refid", None)
            stddev = 0

            # is this an intensity observation or an instrument?
            instrumented = int(netid.lower() not in CIIM_TUPLE)

            station_rows.append(
                (
                    sta_id,
                    network,
                    code,
                    name,
                    lat,
                    lon,
                    elev,
                    vs30,
                    stddev,
                    instrumented,
                    source,
                    refid,
                )
            )

            if not instrumented:
                try:
                    amplitude = float(
                        feature["properties"].get("intensity", np.nan)
                    )
                except (ValueError, TypeError):
                    amplitude = np.nan
                try:
                    stddev = float(
                        feature["properties"].get("intensity_stddev", np.nan)
                    )
                except (ValueError, TypeError):
                    stddev = np.nan
                try:
                    nresp = int(feature["properties"].get("nresp", -1))
                except (ValueError, TypeError):
                    nresp = -1
                if nresp >= 0 and nresp < min_nresp:
                    continue
                flag = feature["properties"]["intensity_flag"]
                if not flag or flag == "":
                    flag = "0"
                amp_rows.append(
                    [sta_id, "MMI", "mmi", "h", amplitude, stddev, flag, nresp]
                )
                imt_set.add("MMI")
                continue

            #
            # Collect the channel names for this station to see if we
            # can resolve the orientation of any of the channels ending
            # with "1" or "2".
            #
            chan_names = []
            for comp in feature["properties"]["channels"]:
                chan_names.append(comp["name"])
            # Some legacy data stupidly names the channel the same as the
            # station name. If there is only one channel, and its name
            # is the station name, we assume it's horizontal and move on.
            if len(chan_names) == 1 and chan_names[0] == name:
                orients = ["H"]
            else:
                orients = _getOrientationSet(chan_names)
            #
            # Now insert the amps into the database
            #
            for ic, comp in enumerate(feature["properties"]["channels"]):
                original_channel = comp["name"]
                orientation = orients[ic]
                for amp in comp["amplitudes"]:
                    if amp["name"].startswith("psa"):
                        imt_type = PSA_TO_SA[amp["name"]]
                    else:
                        imt_type = amp["name"].upper()
                    imt_set.add(imt_type)
                    amp_id = sta_id + "." + imt_type + "." + original_channel
                    if amp_id in amp_set:
                        continue
                    amp_set.add(amp_id)
                    amplitude = amp["value"]
                    if "ln_sigma" in amp:
                        stddev = amp["ln_sigma"]
                    elif "sigma" in amp:
                        stddev = amp["sigma"]
                    else:
                        stddev = 0
                    # We don't care about previous flagging, we'll do our own
                    # flag = amp["flag"]
                    flag = "0"
                    units = amp["units"]
                    if (
                        amplitude == "null"
                        or np.isnan(float(amplitude))
                        or amplitude <= 0
                    ):
                        amplitude = "NULL"
                        flag = "G"
                    elif imt_type == "MMI":
                        pass
                    else:
                        amplitude = convert_units(imt_type, units, amplitude)
                    amp_rows.append(
                        [
                            sta_id,
                            imt_type,
                            original_channel,
                            orientation,
                            amplitude,
                            stddev,
                            flag,
                            -1,
                        ]
                    )
        return (station_rows, amp_rows, reference_rows)

    def _parse_groundpacket_json(self, jsonfile, sta_set, imt_set, amp_set):
        packet = GroundMotionPacket.load_from_json(jsonfile)
        station_rows = []
        amp_rows = []
        reference_rows = []
        for feature in packet.features:
            network_code = feature.properties.network_code
            station_code = feature.properties.station_code
            station_id = f"{network_code}.{station_code}"
            if station_id in sta_set:
                continue
            sta_set.add(station_id)
            station_name = feature.properties.name
            station_lon, station_lat, station_elev = (
                feature.geometry.coordinates
            )
            vs30 = None
            station_rows.append(
                (
                    station_id,
                    network_code,
                    station_code,
                    station_name,
                    station_lat,
                    station_lon,
                    station_elev,
                    vs30,
                    0.0,
                    1,
                    None,
                    None,
                )
            )
            for stream in feature.properties.streams:
                for trace in stream.traces:
                    channel_code = trace.properties.channel_code
                    location_code = trace.properties.location_code
                    if len(location_code.strip()) == 0:
                        location_code = "--"
                    is_channel = True
                    for pattern in NON_CHANNEL_PATTERNS:
                        if re.search(pattern, channel_code) is not None:
                            is_channel = False
                            break
                    if not is_channel:
                        continue

                    if channel_code.lower().startswith("rot"):
                        orient = "H"
                    else:
                        if channel_code[-1].upper() in ORIENTATIONS:
                            orient = ORIENTATIONS[channel_code[-1].upper()]
                        else:
                            orient = None
                    orientation = _getOrientation(channel_code, orient)

                    for metric in trace.metrics:
                        imt_type = metric.properties.name
                        if imt_type not in SUPPORTED_METRICS:
                            continue

                        stddev = 0.0
                        if isinstance(metric.values, float):
                            amplitude = metric.values
                            amplitude = convert_units(
                                imt_type, metric.properties.units, amplitude
                            )
                            flag = 0
                            nresp = -1
                            imt_set.add(imt_type)
                            amp_rows.append(
                                [
                                    station_id,
                                    imt_type,
                                    f"{channel_code}",
                                    orientation,
                                    amplitude,
                                    stddev,
                                    flag,
                                    nresp,
                                ]
                            )
                        else:
                            for idx, amplitude in enumerate(metric.values[0]):
                                amplitude = convert_units(
                                    imt_type,
                                    metric.properties.units,
                                    amplitude,
                                )
                                period = metric.dimensions.axis_values[1][idx]
                                if period >= 1:
                                    imt_type = f"{metric.properties.name}({period:.1f})"
                                else:
                                    imt_type = (
                                        f"{metric.properties.name}({period:g})"
                                    )
                                imt_set.add(imt_type)
                                amp_id = (
                                    f"{network_code}.{station_code}.{imt_type}"
                                    f".{location_code}.{channel_code}"
                                )
                                if amp_id in amp_set:
                                    continue
                                amp_set.add(amp_id)
                                amp_rows.append(
                                    [
                                        station_id,
                                        imt_type,
                                        f"{channel_code}",
                                        orientation,
                                        amplitude,
                                        stddev,
                                        flag,
                                        nresp,
                                    ]
                                )
        return (station_rows, amp_rows, reference_rows)

    def _loadFromJSON(self, jsonfiles, min_nresp):
        """
        Create a StationList object by reading one or more ShakeMap JSON
        data files.

        Args:
            jsonfiles (sequence of str):
                Sequence of ShakeMap JSON data files to read.
            min_nresp (int):
                The minimum number of DYFI responses for an observation to be
                included in the station output.

        Returns:
            nothing: Nothing.
        """

        #
        # Get the station codes for all the stations in the db
        #
        query = "SELECT id FROM station"
        self.cursor.execute(query)
        sta_set = {z[0] for z in self.cursor.fetchall()}

        orig_imt_set = self.getIMTtypes()
        imt_set = orig_imt_set.copy()

        amp_set = set()
        station_rows = []
        amp_rows = []
        ref_rows = []
        for jfile in jsonfiles:
            try:
                tstations, tamps, trefs = self._parse_groundpacket_json(
                    jfile, sta_set, imt_set, amp_set
                )
            except Exception:
                tstations, tamps, trefs = self._parse_shakemap_json(
                    jfile, min_nresp, sta_set, imt_set, amp_set
                )
            station_rows += tstations
            amp_rows += tamps
            ref_rows += trefs

        new_imts = imt_set - orig_imt_set
        if any(new_imts):
            self.cursor.executemany(
                "INSERT INTO imt (imt_type) VALUES (?)", zip(new_imts)
            )
            self.db.commit()
        query = "SELECT imt_type, id FROM imt"
        self.cursor.execute(query)
        imt_hash = dict(self.cursor.fetchall())

        for row in amp_rows:
            row[1] = imt_hash[row[1]]

        self.cursor.executemany(
            "INSERT INTO amp (station_id, imt_id, original_channel, "
            "orientation, amp, stddev, flag, nresp) VALUES "
            "(?, ?, ?, ?, ?, ?, ?, ?)",
            amp_rows,
        )
        self.db.commit()

        self.cursor.executemany(
            "INSERT INTO station (id, network, code, name, lat, lon, "
            "elev, vs30, stddev, instrumented, source, refid) VALUES "
            "(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            station_rows,
        )
        self.db.commit()

        self.cursor.executemany(
            "INSERT INTO reference (shortref, longref, description) VALUES "
            "(?, ?, ?)",
            ref_rows,
        )

    def getGeoJson(self):
        jdict = {"type": "FeatureCollection", "features": []}
        query = "SELECT id, shortref, longref, description FROM reference"
        self.cursor.execute(query)
        rows = self.cursor.fetchall()
        if len(rows):
            refdict = {}
            for row in rows:
                refid = row[0]
                shortref = row[1]
                longref = row[2]
                description = row[3]
                refdict[refid] = {
                    "refid": refid,
                    "long_reference": longref,
                    "short_reference": shortref,
                    "description": description,
                }
            jdict["references"] = refdict
        self.cursor.execute(
            "SELECT id, network, code, name, lat, lon, elev, vs30, "
            "instrumented, source, refid from station"
        )
        sta_rows = self.cursor.fetchall()

        for sta in sta_rows:
            feature = {
                "type": "Feature",
                "id": sta[0],
                "properties": {
                    "code": str(sta[2]),
                    "name": sta[3],
                    "refid": sta[10],
                    "instrumentType": "UNK" if sta[8] == 1 else "OBSERVED",
                    "source": sta[9],
                    "network": sta[1],
                    "commType": "UNK",
                    "location": "",
                    "intensity": None,
                    "intensity_flag": "",
                    "intensity_stddev": None,
                    "pga": None,
                    "pgv": None,
                    "distance": None,
                    "elev": sta[6],
                    "vs30": sta[7],
                    "channels": [],
                },
                "geometry": {"type": "Point", "coordinates": [sta[5], sta[4]]},
            }
            self.cursor.execute(
                "SELECT a.amp, i.imt_type, a.original_channel, "
                "a.flag, a.stddev, a.orientation, a.nresp "
                "FROM amp a, imt i "
                f"WHERE a.station_id = '{str(sta[0])}' "
                "AND a.imt_id = i.id"
            )
            amp_rows = self.cursor.fetchall()
            if sta[8] == 0:
                if len(amp_rows) != 1:
                    logging.warning("Couldn't find intensity for MMI station.")
                    continue
                feature["properties"]["intensity"] = amp_rows[0][0]
                feature["properties"]["intensity_stddev"] = amp_rows[0][4]
                feature["properties"]["intensity_flag"] = amp_rows[0][3]
                feature["properties"]["nresp"] = amp_rows[0][6]
                feature["properties"]["channels"] = []
                jdict["features"].append(feature)
                continue

            channels = {}
            for amp in amp_rows:
                sd_string = "ln_sigma"
                if amp[2] not in channels:
                    channels[amp[2]] = {"name": amp[2], "amplitudes": []}
                if amp[0] == "NULL":
                    value = "null"
                    sigma = "null"
                else:
                    value = amp[0]
                    sigma = float(f"{amp[4]:.4f}")
                if amp[1] == "PGV":
                    if value != "null":
                        value = float(f"{np.exp(value):.4f}")
                    units = "cm/s"
                elif amp[1] == "MMI":
                    if value != "null":
                        value = float(f"{value:.1f}")
                    units = "intensity"
                    sd_string = "sigma"
                else:
                    if value != "null":
                        value = float(f"{np.exp(value) * 100:.4f}")
                    units = "%g"
                aflag = str(amp[3])
                if "I" in aflag and "IncompleteRecord" not in aflag:
                    aflag = aflag.replace("I", "IncompleteRecord")
                if "G" in aflag and "Glitch" not in aflag:
                    aflag = aflag.replace("G", "Glitch")
                if "T" in aflag and "Outlier" not in aflag:
                    aflag = aflag.replace("T", "Outlier")
                if "M" in aflag and "ManuallyFlagged" not in aflag:
                    aflag = aflag.replace("M", "ManuallyFlagged")
                if amp[5] == "U":
                    if aflag == "0":
                        aflag = "UnknownOrientation"
                    else:
                        aflag = str(amp[3]) + ",UnknownOrientation"
                this_amp = {
                    "name": amp[1].lower(),
                    "value": value,
                    "units": units,
                    "flag": aflag,
                    sd_string: sigma,
                }
                channels[amp[2]]["amplitudes"].append(this_amp)
            for channel in channels.values():
                feature["properties"]["channels"].append(channel)
            jdict["features"].append(feature)

        return jdict

    def _loadFromDict(self, stationdict, imtset):
        """
        Internal method to turn the station dictionary created from the
        ShakeMap XML input files into a SQLite database.

        Args:
            stationdictlist (list of stationdicts):
                A list of station dictionaries returned by _filter_station().
            dbfile (string):
                The path to which the SQLite database will be written.

        Returns:
            :class:`StationList` object
        """
        #
        # Get the current IMTs and their IDs and add any new ones
        #
        imts_in_db = self.getIMTtypes()
        new_imts = imtset - imts_in_db
        if any(new_imts):
            self.cursor.executemany(
                "INSERT INTO imt (imt_type) VALUES (?)", zip(new_imts)
            )
            self.db.commit()

        # Now get the updated list of IMTs and their IDs
        query = "SELECT imt_type, id FROM imt"
        self.cursor.execute(query)
        imt_hash = dict(self.cursor.fetchall())

        #
        # Get the station codes for all the stations in the db
        #
        query = "SELECT id FROM station"
        self.cursor.execute(query)
        sta_set = {z[0] for z in self.cursor.fetchall()}

        #
        # Insert any new stations into the station table
        #
        station_rows = []
        for sta_id, station_tpl in stationdict.items():
            if sta_id in sta_set:
                continue
            else:
                sta_set.add(sta_id)
            station_attributes, comp_dict = station_tpl
            lat = station_attributes["lat"]
            lon = station_attributes["lon"]
            code = station_attributes["code"]
            # the attributes dictionary may not have the same
            # netid that we created. Use instead the first part of
            # the station id
            netid = sta_id[0 : sta_id.find(".")]
            network = netid
            name = station_attributes.get("name", None)
            elev = station_attributes.get("elev", None)
            vs30 = station_attributes.get("vs30", None)
            stddev = station_attributes.get("stddev", 0)
            source = station_attributes.get("source", None)
            refid = station_attributes.get("refid", None)

            instrumented = int(network.lower() not in CIIM_TUPLE)

            station_rows.append(
                (
                    sta_id,
                    network,
                    code,
                    name,
                    lat,
                    lon,
                    elev,
                    vs30,
                    stddev,
                    instrumented,
                    source,
                    refid,
                )
            )

        self.cursor.executemany(
            "INSERT INTO station (id, network, code, name, lat, lon, "
            "elev, vs30, stddev, instrumented, source, refid) VALUES "
            "(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            station_rows,
        )
        self.db.commit()

        #
        # Now add the amps, first get the current set so we don't add
        # any duplicates; a unique amp will be (station_id, imt_id,
        # original_channel)
        #
        query = "SELECT station_id, imt_id, original_channel FROM amp"
        self.cursor.execute(query)
        amp_rows = self.cursor.fetchall()
        # Create a unique identifier for each amp so we don't repeat any
        amp_set = set(
            [str(v[0]) + "." + str(v[1]) + "." + str(v[2]) for v in amp_rows]
        )

        #
        # Insert the amps for each component
        #
        amp_rows = []
        for sta_id, station_tpl in stationdict.items():
            station_attributes, comp_dict = station_tpl
            instrumented = int(
                station_attributes["netid"].lower() not in CIIM_TUPLE
            )
            for original_channel, cdict in comp_dict.items():
                pgm_dict = cdict["amps"]
                if len(
                    comp_dict
                ) == 1 and original_channel == station_attributes.get(
                    "name", ""
                ):
                    orientation = "H"
                else:
                    orientation = cdict["attrs"].get("orientation", None)
                    orientation = _getOrientation(
                        original_channel, orientation
                    )
                for imt_type, imt_dict in pgm_dict.items():
                    if (instrumented == 0) and (imt_type != "MMI"):
                        continue
                    imtid = imt_hash[imt_type]
                    amp_id = (
                        str(sta_id)
                        + "."
                        + str(imtid)
                        + "."
                        + str(original_channel)
                    )
                    if amp_id in amp_set:
                        continue
                    amp_set.add(amp_id)
                    amp = imt_dict["value"]
                    units = imt_dict["units"]
                    stddev = imt_dict["stddev"]
                    flag = imt_dict["flag"]
                    nresp = imt_dict.get("nresp", -1)
                    if np.isnan(amp):
                        amp = "NULL"
                        flag = "G"
                    elif imt_type == "MMI":
                        if amp <= 0:
                            amp = "NULL"
                            flag = "G"
                        else:
                            pass
                    elif imt_type == "PGV":
                        if units == "cm/s":
                            if amp <= 0:
                                amp = "NULL"
                                flag = "G"
                            else:
                                amp = np.log(amp)
                        elif units == "ln(cm/s)":
                            pass
                        else:
                            raise ValueError(f"Unknown units {units} in input")
                    else:
                        if units == "%g":
                            if amp <= 0:
                                amp = "NULL"
                                flag = "G"
                            else:
                                amp = np.log(amp / 100.0)
                        elif units == "ln(g)":
                            pass
                        else:
                            raise ValueError(f"Unknown units {units} in input")

                    amp_rows.append(
                        (
                            sta_id,
                            imtid,
                            original_channel,
                            orientation,
                            amp,
                            stddev,
                            flag,
                            nresp,
                        )
                    )

        self.cursor.executemany(
            "INSERT INTO amp (station_id, imt_id, original_channel, "
            "orientation, amp, stddev, flag, nresp) VALUES "
            "(?, ?, ?, ?, ?, ?, ?, ?)",
            amp_rows,
        )
        self.db.commit()
        return

    def getIMTtypes(self):
        """
        Return a set of IMT types found in the database

        Args:
            None

        Returns:
            A set of IMT types
        """
        self.cursor.execute("SELECT imt_type FROM imt")
        return {z[0] for z in self.cursor.fetchall()}

    def getStationDictionary(
        self,
        component="GREATER_OF_TWO_HORIZONTAL",
        instrumented=True,
        min_nresp=1,
    ):
        """
        Return a dictionary of the instrumented or non-instrumented
        stations. The keys describe the parameter, the values are Numpy
        arrays of the parameter in question.

        For the standard set of ShakeMap IMTs (mmi, pga, pgv, psa03, psa10,
        psa30), the keys in the dictionary would be:

        'id', 'network', 'code', 'name', 'lat', 'lon', 'elev', 'vs30',
        'stddev', 'instrumented', 'PGA', 'PGA_sd', 'PGV', 'PGA_sd',
        'SA(0.3)', 'SA(0.3)_sd, 'SA(1.0)', 'SA(1.0)_sd', 'SA(3.0)',
        'SA(3.0)_sd'

        For the non-instrumented dictionary, the keys would be:

        'id', 'network', 'code', 'name', 'lat', 'lon', 'elev', 'vs30',
        'stddev', 'instrumented', 'MMI', 'MMI_sd', 'nresp'

        The **id** column is **network** and **code** concatenated with a
        period (".") between them.

        All ground motion units are natural log units. Distances are in km.

        Args:
            instrumented (bool):
                Set to True if the dictionary is to contain the instrumented
                stations, or to False if the dictionary is to contain the
                non-instrumented (MMI) stations.
            component (str):
                Defines the component of the data returned. The default is
                "GREATER_OF_TWO_HORIZONTAL". Other choices might be "RotD50",
                "RotD100", etc. Invalid components will simply not return any
                IMTs.
            min_nresp (int):
                The minimum number of DYFI responses required to make a valid
                observation.

        Returns:
            dict, set: A dictionary of Numpy arrays, and a set specifying
            the IMTs found in the dictionary.

        Raises:
            TypeError: if "instrumented" argument is not type bool.
        """

        if not isinstance(instrumented, bool):
            raise TypeError(
                "getStationDictionary: the instrumented argument "
                "must be of type bool"
            )
        columns = list(TABLES["station"].keys())
        dstr = ", ".join(columns)
        self.cursor.execute(
            f"SELECT {dstr} FROM station where instrumented = {instrumented}"
        )

        station_rows = self.cursor.fetchall()
        nstation_rows = len(station_rows)
        if not nstation_rows:
            return None, set()
        station_columns = list(zip(*station_rows))
        df = OrderedDict()
        for ic, col in enumerate(columns):
            df[col] = np.array(station_columns[ic])

        myimts = set()
        for imt in self.getIMTtypes():
            if (instrumented and "MMI" in imt) or (
                not instrumented and "MMI" not in imt
            ):
                continue
            df[imt] = np.full(nstation_rows, np.nan)
            df[imt + "_sd"] = np.full(nstation_rows, 0.0)
            if instrumented is False:
                df[imt + "_nresp"] = np.full(nstation_rows, -1, dtype=np.int32)
            myimts.update([imt])

        id_dict = dict(zip(df["id"], range(nstation_rows)))

        #
        # Get all of the unflagged amps with the proper orientation
        #
        self.cursor.execute(
            "SELECT "
            "a.amp, i.imt_type, a.station_id, a.stddev, a.nresp, "
            "a.original_channel "
            "FROM "
            "amp a, station s, imt i "
            "WHERE a.flag = '0' "
            "AND s.id = a.station_id "
            "AND a.imt_id = i.id "
            "AND s.instrumented = ? AND a.orientation NOT IN ('Z', 'U') "
            "AND a.amp IS NOT NULL "
            "AND (a.nresp < 0 OR a.nresp >= ?)",
            (
                instrumented,
                min_nresp,
            ),
        )
        amp_rows = self.cursor.fetchall()

        #
        # Go through all the amps and put them into the data frame
        #
        is_rot_component = component.lower().startswith("rot")
        for this_row in amp_rows:
            #
            # Throw out the components that don't fit the request
            #
            is_rot_row = this_row[5].lower().startswith("rot")
            if is_rot_component and not is_rot_row:
                continue
            component_not_rotd = component in [
                "GREATER_OF_TWO_HORIZONTAL",
                "GEOMETRIC_MEAN",
                "ARITHMETIC_MEAN",
            ]
            amp_is_rotd = this_row[5].lower().startswith("rot")
            if component_not_rotd and amp_is_rotd:
                continue
            #
            # Set the cell to the peak amp
            #
            if component == "GREATER_OF_TWO_HORIZONTAL" or is_rot_component:
                rowidx = id_dict[this_row[2]]
                cval = df[this_row[1]][rowidx]
                amp = this_row[0]
                stddev = this_row[3]
                nresp = this_row[4]
                if np.isnan(cval) or (cval < amp):
                    df[this_row[1]][rowidx] = amp
                    df[this_row[1] + "_sd"][rowidx] = stddev
                    if instrumented is False:
                        df[this_row[1] + "_nresp"][rowidx] = nresp
            else:
                rowidx = id_dict[this_row[2]]
                cval = df[this_row[1]][rowidx]
                amp = this_row[0]
                stddev = this_row[3]
                if np.isnan(cval):
                    df[this_row[1]][rowidx] = amp
                    df[this_row[1] + "_sd"][rowidx] = stddev
                else:
                    if component == "GEOMETRIC_MEAN":
                        df[this_row[1]][rowidx] = np.sqrt(cval * amp)
                    elif component == "ARITHMETIC_MEAN":
                        df[this_row[1]][rowidx] = (cval + amp) / 2.0

        return df, myimts

    def _fixOrientations(self):
        """
        Look for stations with channels that have orientations "1", "2",
        and "Z", and set the "1" and "2" channels to have horizontal
        orientation.
        """
        sql = (
            "SELECT DISTINCT station_id, original_channel "
            "FROM amp "
            "WHERE orientation IN ('U', 'Z') "
            "ORDER BY station_id"
        )
        self.cursor.execute(sql)
        sta_rows = self.cursor.fetchall()
        current_station_id = ""
        channel_list = []
        for row in sta_rows:
            station_id, channel = row
            if current_station_id == "":
                current_station_id = station_id
                channel_list = [channel]
            elif station_id == current_station_id:
                channel_list.append(channel)
            if len(channel_list) == 3:
                ochars = [x[-1].upper() for x in channel_list]
                if "1" in ochars and "2" in ochars and "Z" in ochars:
                    hinds = [
                        i for i, x in enumerate(ochars) if x in ("1", "2")
                    ]
                    sql = (
                        "UPDATE amp "
                        "SET orientation = 'H'"
                        "WHERE station_id = ? "
                        "AND original_channel IN (?, ?)"
                    )
                    self.cursor.execute(
                        sql,
                        (
                            current_station_id,
                            channel_list[hinds[0]],
                            channel_list[hinds[1]],
                        ),
                    )
                    self.db.commit()
                    hinds = [i for i, x in enumerate(ochars) if x == "Z"]
                    sql = (
                        "UPDATE amp "
                        "SET orientation = 'Z' "
                        "WHERE station_id = ? "
                        "AND original_channel = ?"
                    )
                    self.cursor.execute(
                        sql,
                        (
                            current_station_id,
                            channel_list[hinds[0]],
                        ),
                    )
                    self.db.commit()
                # 3 channels, but not 1, 2, Z: bummer
                channel_list = []
            # Channels weren't 3 so we can't fix; bummer
            if current_station_id != station_id:
                current_station_id = station_id
                channel_list = [channel]
        return

    @staticmethod
    def _getGroundMotions(comp, imt_translate):
        """
        Get a dictionary of peak ground motions (values and flags).
        Output keys are one of: [pga,pgv,psa03,psa10,psa30]
        Even if flags are not specified in the input, they will
        be guaranteed to at least have a flag of '0'.
        """
        pgmdict = {}
        imtset = set()
        for pgm in comp:
            if pgm.tag == "#text":
                continue
            key = pgm.tag
            if key not in imt_translate:
                if key in ("acc", "pga"):
                    new_key = "PGA"
                elif key in ("vel", "pgv"):
                    new_key = "PGV"
                elif "mmi" in key:
                    new_key = "MMI"
                elif "psa" in key:
                    pp = get_imt_period(key)
                    new_key = "SA(" + str(pp) + ")"
                else:
                    raise ValueError(f"Unknown amp type in input: {key}")
                imt_translate[key] = new_key
            else:
                new_key = imt_translate[key]
            key = new_key

            if "value" in pgm.attrib:
                try:
                    value = float(pgm.attrib["value"])
                except (ValueError, TypeError):
                    logging.warning(
                        "Unknown value in XML: %s for amp: %s",
                        pgm.attrib["value"],
                        pgm.tag,
                    )
                    continue
            else:
                logging.warning("No value for amp %s", pgm.tag)
                continue
            if "flag" in pgm.attrib and pgm.attrib["flag"] != "":
                flag = pgm.attrib["flag"]
            else:
                flag = "0"
            if "ln_stddev" in pgm.attrib:
                stddev = float(pgm.attrib["ln_stddev"])
            else:
                stddev = 0
            if "units" in pgm.attrib:
                units = pgm.attrib["units"]
            else:
                if key == "PGV":
                    units = "cm/s"
                elif key == "MMI":
                    units = "intensity"
                else:
                    units = "%g"
            pgmdict[key] = {
                "value": value,
                "flag": flag,
                "stddev": stddev,
                "units": units,
            }
            imtset.add(key)
        return pgmdict, imtset, imt_translate

    def _filter_station(self, xmlfile, stationdict, min_nresp):
        """
        Filter individual xmlfile into a stationdict data structure.

        Args:
            xmlfile (string):
                Path to ShakeMap XML input file (or file-like object)
                containing station data.
            stationdict (dict):
                the dictionary of stations that the stations in this file
                should be added to.
            min_nresp (int):
                The minimum number of responses that a DYFI observation needs
                to be included in the processing.

        Returns:
            stationdict data structure
            imts: a set of IMTs that were found in the file
        """
        #
        # Strip off any namespace garbage that is prepended
        # to the tags
        #
        it = ET.iterparse(xmlfile)
        for _, el in it:
            if "}" in el.tag:
                el.tag = el.tag.split("}", 1)[1]
        #
        # Parse the cleaned up xml tree
        #
        imt_translate = {}
        imtset = set()
        root = it.root
        for sl in root.iter("stationlist"):
            if "reference" in sl.attrib:
                longref = sl.attrib["reference"]
                shortref = ""
                description = ""
                query = (
                    "INSERT INTO reference "
                    "(shortref, longref, description) VALUES "
                    f"('{shortref}','{longref}','{description}')"
                )
                self.cursor.execute(query)
                self.db.commit()
                query = (
                    f"SELECT id FROM reference WHERE longref is ('{longref}')"
                )
                self.cursor.execute(query)
                row = self.cursor.fetchall()
                refid = row[0][0]
            else:
                refid = None
            for station in sl:
                if station.tag != "station":
                    continue
                # look at the station attributes to figure out if this is a
                # DYFI-type station or a station with instruments measuring
                # PGA, PGV, etc.
                attributes = station.attrib.copy()
                if "netid" in attributes:
                    netid = attributes["netid"]
                    if len(netid.strip()) == 0:
                        netid = "unknown"
                else:
                    netid = "unknown"
                attributes["netid"] = netid
                attributes["refid"] = refid
                instrumented = int(netid.lower() not in CIIM_TUPLE)

                if "code" not in attributes:
                    logging.warning(
                        "Station does not have station code: skipping"
                    )
                    continue
                code = attributes["code"]
                if code.startswith(netid + "."):
                    sta_id = code
                    code = code.replace(netid + ".", "")
                    attributes["code"] = code
                else:
                    sta_id = netid + "." + code

                if sta_id in stationdict:
                    compdict = stationdict[sta_id][1]
                else:
                    compdict = {}
                for comp in station:
                    if "name" not in comp.attrib:
                        logging.warning(
                            "Unnamed component for station %s; skipping",
                            sta_id,
                        )
                        continue
                    compname = comp.attrib["name"]
                    if "Intensity Questionnaire" in str(compname):
                        compdict["mmi"] = {"amps": {}, "attrs": {}}
                        continue
                    tpgmdict, ims, imt_translate = self._getGroundMotions(
                        comp, imt_translate
                    )
                    if compname in compdict:
                        pgmdict = compdict[compname]["amps"]
                    else:
                        pgmdict = {}
                    pgmdict.update(tpgmdict)
                    # copy the VALUES, not REFERENCES, of the component list
                    # into our growing dictionary
                    compdict[compname] = {
                        "attrs": copy.copy(comp.attrib),
                        "amps": copy.copy(pgmdict),
                    }
                    imtset |= ims
                if ("intensity" in attributes) and (instrumented == 0):
                    if "mmi" not in compdict:
                        compdict["mmi"] = {"amps": {}, "attrs": {}}
                    if "intensity_flag" in attributes:
                        flag = attributes["intensity_flag"]
                    else:
                        flag = "0"
                    if "intensity_stddev" in attributes:
                        stddev = float(attributes["intensity_stddev"])
                    else:
                        stddev = 0
                    if "nresp" in attributes:
                        nresp = int(attributes["nresp"])
                    else:
                        nresp = -1
                    if nresp >= 0 and nresp < min_nresp:
                        continue
                    compdict["mmi"]["amps"]["MMI"] = {
                        "value": float(attributes["intensity"]),
                        "stddev": stddev,
                        "nresp": nresp,
                        "flag": flag,
                        "units": "intensity",
                    }
                    imtset.add("MMI")
                stationdict[sta_id] = (attributes, copy.copy(compdict))
        return stationdict, imtset

    def _createTables(self):
        """
        Build the database tables.
        """
        for table, columns in TABLES.items():
            sql = f"CREATE TABLE {table} ("
            nuggets = []
            for column, ctype in columns.items():
                nuggets.append(f"{column} {ctype}")
            sql += ",".join(nuggets) + ")"
            self.cursor.execute(sql)

        self.db.commit()
        return


def get_imt_period(imt):
    """
    Get the period from a string like psa3p0, psa3.0, or psa30 (the first
    being favored). Return the floating point period.

    Args:
        imt (str): a string starting with "psa" and ending with something
            that can reasonably be converted to a floating point number.

    Returns:
        float: The period of the psa input.

    TODO: Could do a lot more error checking here, but I guess we're
    assuming that the people who send us data aren't idiots.
    """
    # Updated 'psa2p5' style
    p = re.search(r"(?<=psa).*", imt)
    if "p" in p.group(0):
        return float(p.group(0).replace("p", "."))
    # Weird, but we'll allow it psa2.5 style
    if "." in p.group(0):
        return float(p.group(0))
    # Old school psa25 style
    p = re.search(r"(?<=psa)\d+", imt)
    return float(p.group(0)[:-1] + "." + p.group(0)[-1])


def _getOrientation(orig_channel, orient):
    """
    Return a character representing the orientation of a channel.

    Args:
        orig_channel (string):
            String representing the seed channel (e.g. 'HNZ'). The
            final character is assumed to be the (uppercase) orientation.
        orient (str or None):
            Gives the orientation of the channel, overriding channel
            codes that end in numbers. Must be one of 'h' (horizontal)
            or 'v' (vertical), or None if the orientation has not been
            explicitly specified in the "comp" element.

    Returns:
        Character representing the channel orientation. One of 'N',
        'E', 'Z', 'H' (for horizontal), or 'U' (for unknown).
    """
    #
    # Default no-op
    #
    if orient == "H":
        return "H"

    if len(orig_channel.strip()) == 0:
        orientation = "U"  # default when we don't know anything about channel
        return orientation

    if orig_channel == "mmi" or orig_channel == "DERIVED":
        orientation = "H"  # mmi is arbitrarily horizontal
    elif orig_channel[-1].upper() in ("N", "E", "Z"):
        orientation = orig_channel[-1].upper()
    elif orig_channel == "UNK":  # Channel is "UNK"; assume horizontal
        orientation = "H"
    elif orig_channel.upper() == "H1" or orig_channel.upper() == "H2":
        orientation = "H"
    elif orig_channel[-1].isdigit():
        if orient == "h":
            orientation = "H"
        elif orient == "v":
            orientation = "Z"
        else:
            orientation = "U"
    else:
        orientation = "U"  # this is unknown

    return orientation


def _getOrientationSet(chan_names):
    """
    Return a characters representing the orientation of a set of
    channels from a single station.

    Args:
        chan_names (list):
            List of strings representing the seed channels (e.g. 'HNZ').
            The final character is assumed to be the (uppercase)
            orientation.

    Returns:
        Character representing the channel orientation. One of 'N',
        'E', 'Z', 'H' (for horizontal), or 'U' (for unknown).
    """
    if len(chan_names) == 3:
        term_chars = [
            chan_names[0][-1].upper(),
            chan_names[1][-1].upper(),
            chan_names[2][-1].upper(),
        ]
        if "1" in term_chars and "Z" in term_chars:
            orientations = [("Z" if x == "Z" else "H") for x in term_chars]
            return orientations
    orientations = []
    for name in chan_names:
        orientations.append(_getOrientation(name, None))
    return orientations



"""
Perform Monte Carlo simulation of  finite faults and perform a mixed effects
regression to obtain mean ground motions and uncertainties.
"""

import copy
from enum import Enum

import numpy as np
from scipy.stats import norm
import openquake.hazardlib.const as oqconst

from esi_utils_rupture.quad_rupture import QuadRupture
from esi_utils_rupture.point_rupture import PointRupture
from esi_utils_rupture.distance import Distance
from esi_shakelib.sites import Sites
from esi_shakelib.utils.contexts import (
    SitesContext,
    DistancesContext,
    RuptureContext,
)


class AutoName(Enum):
    """
    These enum members hold their name (as a string) as their value.
    For example: DistType.Rjb.value == 'Rjb'
    """

    def __init__(self):
        self._value_ = self.name


class MagScaling(AutoName):
    """
    Magnitude scaling relationship.
       - 'WC94' is for Wells and Coppersmith (1999)
       - 'HB08' is for Hanks and Bakun (2008)
       - 'S14' is for Somerville et al. (2014)
       - 'Sea10_interface' is for Strasser et al. (2010), for interface events
       - 'Sea10_slab' is for Strasser et al. (2010), for intraplate events
    """

    WC94 = ()
    HB08 = ()
    S14 = ()
    Sea10_interface = ()
    Sea10_slab = ()


class Mechanism(AutoName):
    """
    Source mechanism.
       - 'A' for all (mechanism is unknown/unspecified)
       - 'R' for reverse
       - 'N' for normal
       - 'SS' for strike-slip
    """

    ALL = ()
    RS = ()
    NM = ()
    SS = ()


class FFSimmer:
    """
    Class to perform monte carlo simulations of finite faults to derive mean
    ground motions
    """

    def __init__(self, rff, measure="median"):
        """
        Args:
            rff (RandomFiniteFault object):
                An instance of the RandomFiniteFault object.
            measure (str):
                Either "mean" or "median" -- the measure of central tendency
                to report.
        """
        if measure not in ("mean", "median"):
            raise ValueError(f"Unknown measure of central tendency: {measure}")
        self.measure = measure
        self.rff = rff
        self.origin = copy.copy(rff.origin)

    def compute_grid(self, grid, gmpes, myimts):
        """
        Args:
            grid (Grid2D object):
                Basically the vs30grid that we're mapping, used for its
                dimensions.  The site corrections have to be done elsewhere.
            gmpes (dictionary):
                A dictionary of MultiGMPEs to be used for the simulations
                keyed to myimts.
            myimts (list):
                List of OQ IMTs to generate grids and transects for

        Returns:
            dictionary:
                {
                    oqimt: {
                        "mean": 2D Array (the median or mean ground motion
                            values, at a Vs30 of 760 m/s),
                        "transect": Array (a transect of the grid at the repis
                            below),
                        "repi": Array (the Repis corresponding to "transect"
                            above),
                        oqconst.StdDev.TOTAL: Array,
                        (if allowed:)
                        oqconst.StdDev.INTER_EVENT: Array,
                        oqconst.StdDev.INTRA_EVENT: Array,
                    },
                }
        """
        griddict = {}
        hypo = self.origin.getHypo()
        olon = hypo.x
        olat = hypo.y

        if len(gmpes[myimts[0]].DEFINED_FOR_STANDARD_DEVIATION_TYPES) > 1:
            sd_types = [
                oqconst.StdDev.TOTAL,
                oqconst.StdDev.INTER_EVENT,
                oqconst.StdDev.INTRA_EVENT,
            ]
        else:
            sd_types = [
                oqconst.StdDev.TOTAL,
            ]

        gd = grid.getGeoDict()
        # Increase the grid resolution if it is greater the 30 arc-seconds
        delta_y = gd.dy
        if delta_y > (30.0 / 3600.0):
            delta_y = delta_y / 2.0
        # Compute distance to all four corners of grid
        prup = PointRupture(self.origin)
        lons = np.array([gd.xmin, gd.xmin, gd.xmax, gd.xmax])
        lats = np.array([gd.ymin, gd.ymax, gd.ymin, gd.ymax])
        depths = np.zeros_like(lons)
        repi = prup.computeRepi(lons, lats, depths)
        rmax = np.max(repi)

        west = olon
        east = olon
        south = olat
        north = olat + rmax / 111.1
        if north > 90.0:
            north = olat
            south = olat - rmax / 111.1

        sites = Sites.fromBounds(
            west,
            east,
            south,
            north,
            delta_y,
            delta_y,
            defaultVs30=760.0,
            vs30File=None,
            padding=True,
            resample=True,
        )
        (nx, ny) = sites.getNxNy()
        sgd = sites.getVs30Grid().getGeoDict()
        sctx = sites.getSitesContext()
        for k in sctx.__dict__:
            if k == "_slots_":
                continue
            sctx.__dict__[k] = sctx.__dict__[k].flatten()

        nsim = self.rff.get_nsim()
        nimt = len(myimts)
        nporig = nx * ny
        amps = np.zeros((nimt, nsim, nporig))
        sigs = np.zeros((nimt, nsim, nporig))
        if len(sd_types) > 1:
            taus = np.zeros((nimt, nsim, nporig))
            phis = np.zeros((nimt, nsim, nporig))

        for idx in range(nsim):
            rupt = self.rff.retrieve_rupt(idx)
            dist = Distance.fromSites(gmpes[myimts[0]], sites, rupt)
            dctx = dist.getDistanceContext()
            for k in dctx.__dict__:
                if k == "_slots_":
                    continue
                if not isinstance(dctx.__dict__[k], np.ndarray):
                    continue
                dctx.__dict__[k] = dctx.__dict__[k].flatten()
            rctx = rupt.getRuptureContext(gmpes[myimts[0]], (nporig,))
            if idx == 0:
                cols = []
                arrs = []
                for ctx in (sctx, dctx, rctx):
                    for name, val in vars(ctx).items():
                        if ctx == sctx and name in ["lons", "lats"]:
                            continue
                        if isinstance(val, np.ndarray):
                            cols.append(name)
                            arrs.append(
                                np.zeros(nsim * np.size(val), dtype=val.dtype)
                            )
                cx = np.rec.fromarrays(arrs, names=cols)
            for ctx in (sctx, dctx, rctx):
                for name, val in vars(ctx).items():
                    if isinstance(val, np.ndarray):
                        arr = getattr(cx, name)
                        istart = idx * np.size(val)
                        arr[istart : istart + np.size(val)] = val[:]

        for iimt, simt in enumerate(myimts):
            mean, stddevs = gmpes[simt].get_mean_and_stddevs(
                cx, cx, cx, simt, sd_types
            )
            for idx in range(nsim):
                istart = idx * nporig
                amps[iimt, idx, :] = mean[0][istart : istart + nporig]
                sigs[iimt, idx, :] = stddevs[0][istart : istart + nporig]
                if len(sd_types) > 1:
                    taus[iimt, idx, :] = stddevs[1][istart : istart + nporig]
                    phis[iimt, idx, :] = stddevs[2][istart : istart + nporig]

        slat = np.linspace(sgd.ymin, sgd.ymax, sgd.ny)
        repi = prup.computeRepi(
            np.full_like(slat, sgd.xmin),
            slat,
            np.zeros_like(slat),
        )

        gd = grid.getGeoDict()
        glon, glat = np.meshgrid(
            np.linspace(gd.xmin, gd.xmax, gd.nx),
            np.linspace(gd.ymax, gd.ymin, gd.ny),
        )
        gdepth = np.zeros_like(glon)
        grepi = rupt.computeRepi(glon, glat, gdepth)

        # Compute Mean, Sigma, Phi, and Tau for each IMT
        for iimt in range(nimt):
            if self.measure == "median":
                mean_amps = np.median(amps[iimt], axis=0)
            else:
                mean_amps = np.mean(amps[iimt], axis=0)
            residuals = amps[iimt] - mean_amps
            extra_vars = np.var(amps[iimt], axis=0)

            try:
                avg_extra_tau = np.var(np.mean(residuals, axis=1))
                avg_extra_phi = np.mean(np.var(residuals, axis=1))
            except RuntimeWarning:
                # Don't have an answer, so just guess
                print("Warning in FFSimmer: can't compute MER stats")
                print("Using defaults")
                avg_extra_tau = 0.20
                avg_extra_phi = 0.80

            if (
                np.isnan(avg_extra_tau)
                or np.isnan(avg_extra_phi)
                or avg_extra_tau < 1e-8
                or avg_extra_phi < 1e-8
            ):
                print("mixedlm.fit produced unusable results")
                avg_extra_tau = 0.20
                avg_extra_phi = 0.80

            avg_tau_wgt = avg_extra_tau / (avg_extra_tau + avg_extra_phi)
            avg_phi_wgt = 1.0 - avg_tau_wgt
            extra_taus = avg_tau_wgt * extra_vars
            extra_phis = avg_phi_wgt * extra_vars

            # Take the mean over the realizations
            mean_sigs = np.sqrt(np.mean(sigs[iimt] ** 2, axis=0) + extra_vars)
            if len(sd_types) > 1:
                mean_taus = np.sqrt(
                    np.mean(taus[iimt] ** 2, axis=0) + extra_taus
                )
                mean_phis = np.sqrt(
                    np.mean(phis[iimt] ** 2, axis=0) + extra_phis
                )

            griddict[myimts[iimt]] = {
                "mean": np.interp(grepi, repi, np.flip(mean_amps)),
                oqconst.StdDev.TOTAL: np.interp(
                    grepi, repi, np.flip(mean_sigs)
                ),
                "transect": np.flip(mean_amps),
                "repi": repi,
            }
            if len(sd_types) > 1:
                griddict[myimts[iimt]][oqconst.StdDev.INTER_EVENT] = np.interp(
                    grepi, repi, np.flip(mean_taus)
                )
                griddict[myimts[iimt]][oqconst.StdDev.INTRA_EVENT] = np.interp(
                    grepi, repi, np.flip(mean_phis)
                )

        return griddict

    def compute_points(
        self, sctx, gmpes, myimts, rock_vs30=760.0, soil_vs30=180.0
    ):
        """
        Compute median (or mean) ground motions for a set of sites defined in
        "sctx" for the imts in the list "myimts".

        Args:
            sctx: (Sites object)
                A sites context object set up for the points being output
            gmpes: (dictionary)
                A dictionary (keyed to myimts) of gmpes to perform the
                ground-motion realizations.
            myimts (list)
                List of OQ IMTs to generate rock and site amplified values for

        Returns:
            dictionary:
                {
                    oqimt: {
                        "site": The sites at their actual Vs30
                            {
                                "mean": Array,
                                oqconst.StdDev.TOTAL: Array,
                                (if allowed:)
                                oqconst.StdDev.INTER_EVENT: Array,
                                oqconst.StdDev.INTRA_EVENT: Array,
                            }
                        "rock": The sites at a reference rock vs30 of 760 m/s
                        "soil": The sites at a reference soil vs30 of 180 m/s
                    },
                }
        """
        outdict = {}

        if len(gmpes[myimts[0]].DEFINED_FOR_STANDARD_DEVIATION_TYPES) > 1:
            sd_types = [
                oqconst.StdDev.TOTAL,
                oqconst.StdDev.INTER_EVENT,
                oqconst.StdDev.INTRA_EVENT,
            ]
        else:
            sd_types = [
                oqconst.StdDev.TOTAL,
            ]

        sctx_rock = copy.deepcopy(sctx)
        sctx_rock.vs30 = np.full_like(sctx.vs30, rock_vs30)
        sctx_soil = copy.deepcopy(sctx)
        sctx_soil.vs30 = np.full_like(sctx.vs30, soil_vs30)

        nporig = np.size(sctx.lons)
        sx = copy.deepcopy(sctx)
        for k in sctx.__dict__:
            if k == "_slots_":
                continue
            sx.__dict__[k] = np.hstack(
                (
                    sctx.__dict__[k],
                    sctx_rock.__dict__[k],
                    sctx_soil.__dict__[k],
                )
            )
        # sx.sids = np.array(list(range(np.size(sx.lons))))
        depths = np.zeros_like(sx.lons)

        nsd = len(sd_types)
        ctxt_names = ("site", "soil", "rock")
        ctxts = (sctx, sctx_soil, sctx_rock)
        nctx = len(ctxts)

        nimt = len(myimts)
        nsim = self.rff.get_nsim()

        amps = np.zeros((nctx, nimt, nsim, nporig))
        sigs = np.zeros((nctx, nimt, nsim, nporig))
        if nsd > 1:
            taus = np.zeros((nctx, nimt, nsim, nporig))
            phis = np.zeros((nctx, nimt, nsim, nporig))

        llons = sx.lons.flatten()
        llats = sx.lats.flatten()
        ddepths = depths.flatten()
        for idx in range(nsim):
            rupt = self.rff.retrieve_rupt(idx)
            dist_obj = Distance(
                gmpes[myimts[0]],
                llons,
                llats,
                ddepths,
                rupt,
            )
            dctx = dist_obj.getDistanceContext()
            rctx = rupt.getRuptureContext(gmpes[myimts[0]], shape=llons.shape)
            if idx == 0:
                cols = []
                arrs = []
                for ctx in (sx, dctx, rctx):
                    for name, val in vars(ctx).items():
                        if ctx == sx and name in ["lons", "lats"]:
                            continue
                        if isinstance(val, np.ndarray):
                            cols.append(name)
                            arrs.append(
                                np.zeros(nsim * np.size(val), dtype=val.dtype)
                            )
                cx = np.rec.fromarrays(arrs, names=cols)
            for ctx in (sx, dctx, rctx):
                for name, val in vars(ctx).items():
                    if ctx == sx and name in ["lons", "lats"]:
                        continue
                    if isinstance(val, np.ndarray):
                        arr = getattr(cx, name)
                        istart = idx * np.size(val)
                        arr[istart : istart + np.size(val)] = val[:]

        for ix, oqimt in enumerate(myimts):
            mean, stddevs = gmpes[oqimt].get_mean_and_stddevs(
                cx, cx, cx, [oqimt], sd_types
            )
            for idx in range(nsim):
                istart = idx * nctx * nporig
                for ictx in range(nctx):
                    amps[ictx, ix, idx, :] = mean[0][
                        istart + ictx * nporig : istart + (ictx + 1) * nporig
                    ]
                    sigs[ictx, ix, idx, :] = stddevs[0][
                        istart + ictx * nporig : istart + (ictx + 1) * nporig
                    ]
                    if nsd > 1:
                        taus[ictx, ix, idx, :] = stddevs[1][
                            istart
                            + ictx * nporig : istart
                            + (ictx + 1) * nporig
                        ]
                        phis[ictx, ix, idx, :] = stddevs[2][
                            istart
                            + ictx * nporig : istart
                            + (ictx + 1) * nporig
                        ]

        # Compute Mean, Sigma, Phi, and Tau for each IMT
        for ix in range(nimt):
            outdict[myimts[ix]] = {}
            for ictx, ctx_name in enumerate(ctxt_names):
                # Take the mean over the realizations
                if self.measure == "median":
                    mean_amps = np.median(amps[ictx, ix], axis=0)
                else:
                    mean_amps = np.mean(amps[ictx, ix], axis=0)
                # Compute the residuals for the simulations
                residuals = amps[ictx, ix] - mean_amps
                # Find the variance over the realizations, this is the
                # variance we have to account for
                extra_vars = np.var(amps[ictx, ix], axis=0)

                try:
                    avg_extra_tau = np.var(np.mean(residuals, axis=1))
                    avg_extra_phi = np.mean(np.var(residuals, axis=1))
                except RuntimeWarning:
                    # Don't have an answer, so just guess
                    print("Warning in FFSimmer: can't compute MER stats")
                    print("Using defaults")
                    avg_extra_tau = 0.20
                    avg_extra_phi = 0.80

                if (
                    np.isnan(avg_extra_tau)
                    or np.isnan(avg_extra_phi)
                    or avg_extra_tau < 1e-8
                    or avg_extra_phi < 1e-8
                ):
                    print("mixedlm.fit produced unusable results")
                    avg_extra_tau = 0.20
                    avg_extra_phi = 0.80

                avg_tau_wgt = avg_extra_tau / (avg_extra_tau + avg_extra_phi)
                avg_phi_wgt = 1.0 - avg_tau_wgt
                extra_taus = avg_tau_wgt * extra_vars
                extra_phis = avg_phi_wgt * extra_vars

                mean_sigs = np.sqrt(
                    np.mean(sigs[ictx, ix] ** 2, axis=0) + extra_vars
                )
                if len(sd_types) > 1:
                    mean_taus = np.sqrt(
                        np.mean(taus[ictx, ix] ** 2, axis=0) + extra_taus
                    )
                    mean_phis = np.sqrt(
                        np.mean(phis[ictx, ix] ** 2, axis=0) + extra_phis
                    )

                outdict[myimts[ix]].update(
                    {
                        ctx_name: {
                            "mean": mean_amps.reshape(sctx.lons.shape),
                            oqconst.StdDev.TOTAL: mean_sigs.reshape(
                                sctx.lons.shape
                            ),
                        }
                    }
                )
                if len(sd_types) > 1:
                    outdict[myimts[ix]][ctx_name][
                        oqconst.StdDev.INTER_EVENT
                    ] = mean_taus.reshape(sctx.lons.shape)
                    outdict[myimts[ix]][ctx_name][
                        oqconst.StdDev.INTRA_EVENT
                    ] = mean_phis.reshape(sctx.lons.shape)
        return outdict

    def true_grid(self, grid, gmpes, myimts):
        """
        Perform the simulations on a full grid, including the actual Vs30
        Args:
            grid (Grid2D object):
                The vs30grid that we're mapping.
            gmpes (dictionary):
                A dictionary of MultiGMPEs to be used for the simulations
                keyed to myimts.
            myimts (list):
                List of OQ IMTs to generate grids and transects for

        Returns:
            dictionary:
                {
                    oqimt: {
                        "mean": 2D Array (the median or mean ground motion
                            values, at a Vs30 of 760 m/s),
                        oqconst.StdDev.TOTAL: Array,
                        (if allowed:)
                        oqconst.StdDev.INTER_EVENT: Array,
                        oqconst.StdDev.INTRA_EVENT: Array,
                    },
                }
        """
        griddict = {}
        if len(gmpes[myimts[0]].DEFINED_FOR_STANDARD_DEVIATION_TYPES) > 1:
            sd_types = [
                oqconst.StdDev.TOTAL,
                oqconst.StdDev.INTER_EVENT,
                oqconst.StdDev.INTRA_EVENT,
            ]
        else:
            sd_types = [
                oqconst.StdDev.TOTAL,
            ]
        sites = Sites(grid, defaultVs30=760.0)
        (nx, ny) = sites.getNxNy()
        sctx = sites.getSitesContext()
        for k in sctx.__dict__:
            if k == "_slots_":
                continue
            sctx.__dict__[k] = sctx.__dict__[k].flatten()

        nsim = self.rff.get_nsim()
        nimt = len(myimts)
        nporig = nx * ny
        amps = np.zeros((nimt, nsim, nporig))
        sigs = np.zeros((nimt, nsim, nporig))
        if len(sd_types) > 1:
            taus = np.zeros((nimt, nsim, nporig))
            phis = np.zeros((nimt, nsim, nporig))

        for idx in range(nsim):
            rupt = self.rff.retrieve_rupt(idx)
            dist = Distance.fromSites(gmpes[myimts[0]], sites, rupt)
            dctx = dist.getDistanceContext()
            for k in dctx.__dict__:
                if k == "_slots_":
                    continue
                if not isinstance(dctx.__dict__[k], np.ndarray):
                    continue
                dctx.__dict__[k] = dctx.__dict__[k].flatten()
            rctx = rupt.getRuptureContext(gmpes[myimts[0]], (nporig,))

            for iimt, simt in enumerate(myimts):
                mean, stddevs = gmpes[simt].get_mean_and_stddevs(
                    sctx, rctx, dctx, simt, sd_types
                )
                amps[iimt, idx, :] = mean[0]
                sigs[iimt, idx, :] = stddevs[0]
                if len(sd_types) > 1:
                    taus[iimt, idx, :] = stddevs[1]
                    phis[iimt, idx, :] = stddevs[2]

        # Compute Mean, Sigma, Phi, and Tau for each IMT
        for iimt in range(nimt):
            if self.measure == "median":
                mean_amps = np.median(amps[iimt], axis=0)
            else:
                mean_amps = np.mean(amps[iimt], axis=0)
            residuals = amps[iimt] - mean_amps
            extra_vars = np.var(amps[iimt], axis=0)

            try:
                avg_extra_tau = np.var(np.mean(residuals, axis=1))
                avg_extra_phi = np.mean(np.var(residuals, axis=1))
            except RuntimeWarning:
                # Don't have an answer, so just guess
                print("Warning in FFSimmer: can't compute MER stats")
                print("Using defaults")
                avg_extra_tau = 0.20
                avg_extra_phi = 0.80

            if (
                np.isnan(avg_extra_tau)
                or np.isnan(avg_extra_phi)
                or avg_extra_tau < 1e-8
                or avg_extra_phi < 1e-8
            ):
                print("mixedlm.fit produced unusable results")
                avg_extra_tau = 0.20
                avg_extra_phi = 0.80

            avg_tau_wgt = avg_extra_tau / (avg_extra_tau + avg_extra_phi)
            avg_phi_wgt = 1.0 - avg_tau_wgt
            extra_taus = avg_tau_wgt * extra_vars
            extra_phis = avg_phi_wgt * extra_vars

            # Take the mean over the realizations
            mean_sigs = np.sqrt(np.mean(sigs[iimt] ** 2, axis=0) + extra_vars)
            if len(sd_types) > 1:
                mean_taus = np.sqrt(
                    np.mean(taus[iimt] ** 2, axis=0) + extra_taus
                )
                mean_phis = np.sqrt(
                    np.mean(phis[iimt] ** 2, axis=0) + extra_phis
                )

            griddict[myimts[iimt]] = {
                "mean": mean_amps.reshape((ny, nx)),
                oqconst.StdDev.TOTAL: mean_sigs.reshape((ny, nx)),
            }
            if len(sd_types) > 1:
                griddict[myimts[iimt]][oqconst.StdDev.INTER_EVENT] = (
                    mean_taus.reshape((ny, nx))
                )
                griddict[myimts[iimt]][oqconst.StdDev.INTRA_EVENT] = (
                    mean_phis.reshape((ny, nx))
                )

        return griddict


class RandomFiniteFault(object):
    """
    Class to dole out random finite faults efficiently.
    """

    def __init__(
        self,
        origin,
        nsim,
        min_strike=-180.0,
        max_strike=180.0,
        min_dip=None,
        max_dip=None,
        dx_min_frac=0.0,
        dx_max_frac=1.0,
        delta_depth=20.0,
        ztor=-1,
        seed=None,
    ):
        """
        Args:
            origin (Origin object): The origin of the earthquake. Can include
                attributes "tectonic_region" ("Active", "Stable", "Subduction"
                are recognized) and "mech" ("ALL", "NM" (normal), "RS"
                (reverse), "SS" (strike-slip)).
            nsim (int): The number of simulations to run.
            min_strike/max_strike (float/float): Limit the range of the strike
                (in degrees) of realized faults. The default -180/180 allows
                any strike.
            min_dip/max_dip (float/float): Limit the range of the dip (in
                degrees) of the realized faults. The default is to use the
                "mech" of the origin to limit the range or, if none, use
                the "ALL" mech which defaults to 35/90.
            dx_min_frac/dx_max_frac (float/float): Limit the range of the
                along-strike placement of the hypocenter to from
                dx_min_frac * length to dx_max_frac * length. (default
                0.0/1.0)
            delta_depth (float): The depth range of the hypocenter is
                uniformly drawn from depth - delta_depth to
                depth + delta_depth (kilometers) (Default: 20.0)
            ztor (float): Fix the top of rupture to a specific depth
                (kilometers, positive down). Negative values allow the
                TOR to float. (Default -1.)
            seed (int): A seed for the random number generator (for testing
                only).
        """
        self.origin = origin
        self.nsim = nsim
        self.rupts = []
        self.rng = np.random.default_rng(seed)
        self.initialize(
            min_strike,
            max_strike,
            min_dip,
            max_dip,
            dx_min_frac,
            dx_max_frac,
            delta_depth,
            ztor,
        )

    def initialize(
        self,
        min_strike,
        max_strike,
        min_dip,
        max_dip,
        dx_min_frac,
        dx_max_frac,
        delta_depth,
        ztor,
    ):
        """
        Initialize the random arrays
        """
        (mscale, aspect, mindip_deg, maxdip_deg, szdepth, smech) = (
            get_scaling_params(self.origin)
        )
        _, _, _, _, area_mean, area_std = dimensions_from_magnitude(
            self.origin.mag, mscale, 1, 0, smech
        )
        hypo = self.origin.getHypo()
        depth = hypo.z
        if depth > szdepth:
            szdepth = depth
        depth = self.rng.uniform(
            np.maximum(depth - delta_depth, 0),
            np.minimum(depth + delta_depth, szdepth),
            self.nsim,
        )
        area = 10 ** self.rng.normal(
            np.log10(area_mean[0]), area_std, self.nsim
        )
        width = np.sqrt(area / aspect)
        if min_dip is not None:
            mindip_deg = min_dip
            maxdip_deg = max_dip
        dip = np.radians(self.rng.uniform(mindip_deg, maxdip_deg, self.nsim))
        # Need to apply a max to dy that might be less than width based on the
        # eq depth.
        # height = vertical projection of plane
        # height = width * sin(dip)
        #  ------------------
        #  _____ztor       |
        #  \dip    |       |
        #   \      |       |
        #    \     |       |
        #   dy\    |height |
        #      *- -|- - - - depth
        #       \  |
        #        \ |
        #    width\|
        #
        #  ---------------- szdepth
        #
        height = width * np.sin(dip)
        # Don't allow the rupture to extend below the seismogenic depth
        height[height > szdepth] = szdepth
        width = height / np.sin(dip)
        # Limit the rupture to stay between the surface and the szdepth
        # but allow it to float in between
        if ztor < 0:
            dy = self.rng.uniform(
                np.maximum(0, width - (szdepth - depth) / np.sin(dip)),
                np.minimum(width, width - (height - depth) / np.sin(dip)),
            )
        else:
            # Fix ztor to a specific depth. This is a weak contraint: there is
            # no guarantee that for a given depth, dip, and width that the TOR
            # will reach ztor
            dy = (depth - ztor) / np.sin(dip)
            dy[dy > width] = width[dy > width]

        length = area / width
        dx = self.rng.uniform(dx_min_frac * length, dx_max_frac * length)
        strike = self.rng.uniform(min_strike, max_strike, self.nsim)
        hx = np.array([hypo.x])
        hy = np.array([hypo.y])

        for idx in range(self.nsim):
            self.rupts.append(
                QuadRupture.fromOrientation(
                    hx,
                    hy,
                    np.array([depth[idx]]),
                    [dx[idx]],
                    [dy[idx]],
                    [length[idx]],
                    [width[idx]],
                    [strike[idx]],
                    [np.degrees(dip[idx])],
                    self.origin,
                )
            )

    def retrieve_rupt(self, isim):
        """
        Retrieve one fault.
        """
        if isim > self.nsim:
            raise IndexError(
                f"RandomFiniteFault.retrieve_rupt {isim=} exceeds "
                "maximum {self.nsim}"
            )
        return self.rupts[isim]

    def get_nsim(self):
        """
        Get the number of simulated faults that this object holds.
        """
        return self.nsim


def get_scaling_params(origin):
    """

    Args:
        origin:
            An Origin object.
    """

    mech = getattr(origin, "mech", "ALL")
    if not hasattr(origin, "tectonic_region"):
        mscale = MagScaling.WC94
        smech = Mechanism.ALL
        mindip_deg = 35.0
        maxdip_deg = 90.0
        aspect = 1.7
        szdepth = 25.0
    elif origin.tectonic_region == "Active":
        mscale = MagScaling.HB08
        aspect = 1.7
        szdepth = 25.0
        if mech == "ALL":
            # HB08 doesn't have an 'ALL' mechanism, so use WC94
            mscale = MagScaling.WC94
            smech = Mechanism.ALL
            mindip_deg = 35.0
            maxdip_deg = 90.0
        elif mech == "RS":
            smech = Mechanism.RS
            mindip_deg = 35.0
            maxdip_deg = 50.0
        elif mech == "NM":
            smech = Mechanism.NM
            mindip_deg = 40.0
            maxdip_deg = 60.0
        elif mech == "SS":
            smech = Mechanism.SS
            mindip_deg = 75.0
            maxdip_deg = 90.0
    elif origin.tectonic_region == "Stable":
        mscale = MagScaling.S14
        aspect = 1.0
        szdepth = 20.0
        if mech == "ALL":
            smech = Mechanism.ALL
            mindip_deg = 35.0
            maxdip_deg = 90.0
        elif mech == "RS":
            smech = Mechanism.RS
            mindip_deg = 35.0
            maxdip_deg = 60.0
        elif mech == "NM":
            smech = Mechanism.NM
            mindip_deg = 40.0
            maxdip_deg = 60.0
        elif mech == "SS":
            smech = Mechanism.SS
            mindip_deg = 60.0
            maxdip_deg = 90.0
    elif origin.tectonic_region == "Subduction":
        aspect = 1.7
        szdepth = np.inf
        if mech == "ALL":
            mscale = MagScaling.Sea10_interface
            smech = Mechanism.ALL
            mindip_deg = 35.0
            maxdip_deg = 90.0
        elif mech == "RS":
            mscale = MagScaling.Sea10_interface
            smech = Mechanism.RS
            mindip_deg = 35.0
            maxdip_deg = 50.0
        elif mech == "NM":
            mscale = MagScaling.Sea10_slab
            smech = Mechanism.NM
            mindip_deg = 40.0
            maxdip_deg = 60.0
        elif mech == "SS":
            mscale = MagScaling.Sea10_slab
            smech = Mechanism.SS
            mindip_deg = 75.0
            maxdip_deg = 90.0
    else:
        print(
            "Unsupported tectonic region; using coefficients for unknown"
            "tectonic region."
        )
        mscale = MagScaling.WC94
        smech = Mechanism.ALL
        aspect = 1.7
        szdepth = 25.0
        mindip_deg = 35.0
        maxdip_deg = 90.0

    return (mscale, aspect, mindip_deg, maxdip_deg, szdepth, smech)


def dimensions_from_magnitude(
    mag, rup_dim_model, neps, trunc, mech=Mechanism.ALL
):
    """
    Compute dimensions of rupture from magnitude for a specified
    magnitude scaling relation.

    Args:
        mag (float): Magnitude.
        rup_dim_model (MagScaling enum): Specifies the model for compputing the
            rupture dimensions from magnitude.
        neps (int): The number of steps to integrate from -trunc to +trunc.
            Larger numbers increase the accuracy of the result, but take
            longer to run.
        trunc (float): For the integration in area (or length and width), trunc
            is the truncation of the normal distribution (in units of sigma).
        mech (Mechanism enum): Optional string indicating earthquake
            mechanism, used by some of the models.

    Returns:
        tuple: A tuple containing the following, noting that some of these will
        be empty if the selected model does not provide them:

                - length: rupture length (km).
                - sig_length: standard deviation of rupture length.
                - W: rupture width (km).
                - sigw: standard devation of rupture width.
                - A: rupture area (km).
                - siga: standard deivaiton of rupture area.

    """
    epsmid, _, _ = compute_epsilon(neps, trunc)
    if not isinstance(rup_dim_model, MagScaling):
        raise TypeError("rup_dim_model must be of type MagScaling")
    if not isinstance(mech, Mechanism):
        raise TypeError("mech must be of type Mechanism")

    if rup_dim_model is MagScaling.WC94:
        # Use mech to get either M-A or (M-W) and (M-R) from Wells and
        # Coppersmith.
        if mech is Mechanism.SS:
            sig_length = 0.15
            length = 10 ** (-2.57 + 0.62 * mag + sig_length * epsmid)
            sig_width = 0.14
            width = 10 ** (-0.76 + 0.27 * mag + sig_width * epsmid)
            sig_area = 0.22
            area = 10 ** (-3.42 + 0.90 * mag + sig_area * epsmid)
        elif mech is Mechanism.RS:
            sig_length = 0.16
            length = 10 ** (-2.42 + 0.58 * mag + sig_length * epsmid)
            sig_width = 0.15
            width = 10 ** (-1.61 + 0.41 * mag + sig_width * epsmid)
            sig_area = 0.26
            area = 10 ** (-3.99 + 0.98 * mag + sig_area * epsmid)
        elif mech is Mechanism.NM:
            sig_length = 0.17
            length = 10 ** (-1.88 + 0.50 * mag + sig_length * epsmid)
            sig_width = 0.12
            width = 10 ** (-1.14 + 0.35 * mag + sig_width * epsmid)
            sig_area = 0.22
            area = 10 ** (-2.78 + 0.82 * mag + sig_area * epsmid)
        elif mech is Mechanism.ALL:
            sig_length = 0.16
            length = 10 ** (-2.44 + 0.59 * mag + sig_length * epsmid)
            sig_width = 0.15
            width = 10 ** (-1.01 + 0.32 * mag + sig_width * epsmid)
            sig_area = 0.24
            area = 10 ** (-3.49 + 0.91 * mag + sig_area * epsmid)
        else:
            raise TypeError("Unsupported value of 'mech'")
    elif rup_dim_model is MagScaling.S14:
        # Somerville (2014) model:
        #     - No length or width
        #     - No mechanism dependence
        sig_area = 0.3
        area = 10 ** (mag - 4.25 + sig_area * epsmid)
        length = None
        sig_length = None
        width = None
        sig_width = None
    elif rup_dim_model == MagScaling.HB08:
        # Hanks and Bakun (2008)
        # These are the equations reported in the paper:
        #     M =       log10(A) + 3.98   for A <= 537 km^2 w/ se=0.03
        #     M = 4/3 * log10(A) + 3.07   for A >  537 km^2 w/ se=0.04
        # Using them is not so straight-forward beacuse we need to compute
        # the area from magnitude. Of course, this gives a different result
        # than if the equations were regressed for A as a function of M,
        # although since the equations were derived theoretically, this may
        # not be so bad.
        #
        # The inverted equation is simple enough:
        #     log10(A) =      M - 3.98    for M <= 6.71
        #     log10(A) = 3/4*(M - 3.07))  for M > 6.71
        #
        # The standard deviations are a little trickier.
        # First, convert standard errors of M to standard deviations of M:
        # (by my count, n=62 for A<=537, and n=28 for A>537)
        #     0.03*sqrt(62) = 0.236       for M <= 6.71
        #     0.04*sqrt(28) = 0.212       for M > 6.71
        # And convert to standard deviations of log(A) using the partial
        # derivatives
        #     dM/d(log10(A)) = 1          for M <= 6.71
        #     dM/d(log10(A)) = 3/4        for M >  6.71
        # So
        #     0.236*1   = 0.236 (pretty close to WC94)
        #     0.212*3/4 = 0.159 (seems a bit low...)
        if mag > 6.71:
            sig_area = 0.236
            area = 10 ** (3 / 4 * (mag - 3.07) + sig_area * epsmid)
        else:
            sig_area = 0.159
            area = 10 ** ((mag - 3.98) + sig_area * epsmid)
        length = None
        sig_length = None
        width = None
        sig_width = None
    elif rup_dim_model == MagScaling.Sea10_interface:
        # Strasser et al. (2010), coefficients for interface events
        sig_length = 0.18
        length = 10 ** (-2.477 + 0.585 * mag + sig_length * epsmid)
        sig_width = 0.173
        width = 10 ** (-0.882 + 0.351 * mag + sig_width * epsmid)
        sig_area = 0.304
        area = 10 ** (-3.49 + 0.952 * mag + sig_area * epsmid)
    elif rup_dim_model == MagScaling.Sea10_slab:
        # Strasser et al. (2010), coefficients for slab events
        sig_length = 0.146
        length = 10 ** (-2.35 + 0.562 * mag + sig_length * epsmid)
        sig_width = 0.067
        width = 10 ** (-1.058 + 0.356 * mag + sig_width * epsmid)
        sig_area = 0.184
        area = 10 ** (-3.225 + 0.89 * mag + sig_area * epsmid)
    else:
        raise TypeError("Unsupported value of 'rup_dim_model'")
    return length, sig_length, width, sig_width, area, sig_area


def compute_epsilon(neps, trunc):
    """
    Compute midpoints and probabilities of epsilon bins.

    Args:
        neps (int): The number of steps to integrate from -trunc to +trunc.
            Larger numbers increase the accuracy of the result, but take
            longer to run.
        trunc (float): For the integration in area (or length and width), trunc
            is the truncation of the normal distribution (in units of sigma).
    Returns:
        tuple: epsilon midpoints, their probabilities, bin width.
    """
    # Need to assume a truncation level for normal distribution
    eps = np.linspace(-trunc, trunc, neps + 1)
    epsmid = 0.5 * (eps[1:] + eps[:-1])
    peps = norm.cdf(eps[1:]) - norm.cdf(eps[:-1])

    # define delta epsilons to normalize probabilities
    d_eps = 2 * trunc / neps
    epsfac = np.trapz(peps, dx=d_eps)

    if neps > 1:
        peps /= epsfac
    return epsmid, peps, d_eps


# import concurrent.futures
#        with concurrent.futures.ProcessPoolExecutor() as executor:
#            futures = {
#                executor.submit(call_gmpe, ix, gmpes, oqimt, cx, sd_types)
#                for ix, oqimt in enumerate(myimts)
#            }
#            for fut in concurrent.futures.as_completed(futures):
#                (ix, mean, stddevs) = fut.result()
#
# def call_gmpe(ix, gmpes, oqimt, cx, sd_types):
#
#    mean, stddevs = gmpes[oqimt].get_mean_and_stddevs(
#        cx, cx, cx, [oqimt], sd_types
#    )
#    return (ix, mean, stddevs)



import numpy as np

from esi_shakelib.correlation.ccf_base import CrossCorrelationBase


class DummyCorrelation(CrossCorrelationBase):
    """
    Simplified correlation module for testing purposes. Should not be used
    in productions runs as it does not produce valid correlations.
    """

    def __init__(self, periods):
        """
        Initialize the cross-correlation object.

        Args:
            periods (ndarray): An array of periods that will be requested
                from the function. Values must be in the range [0.01, 10.0],
                and must me sorted from smallest to largest.

        Returns:
            An instance of :class:`DummyCorrelation`.
        """

        if np.any(periods < 0.01):
            raise ValueError("The periods must be greater or equal to 0.01s")
        if np.any(periods > 10):
            raise ValueError("The periods must be less or equal to 10s")

        self.periods = periods.copy()

    def getCorrelation(self, ix1, ix2, h):
        """
        Compute the correlation between two periods and a separation distance
        of h km. The result returned is::

          rho = T1/T2 * exp(-h/10)

        where rho is the correlation, T1 is the smaller period, T2 is the
        larger period, and h is the distance between the points of interest.

        The index arrays (ix1 and ix2) and h array must have the same
        dimensions. The
        indices may be equal, and there is no restriction on which one is
        larger. The indices refer to periods in the 'period' argument to the
        class constructor.

        Args:
            ix1 (ndarray):
                The indices of the first period of interest.
            ix2 (ndarrays):
                The indices of the second period of interest.
            h (ndarray):
                The separation distance between two sites (units of km).
                The result is placed in h.

        Returns:
            h (ndarray): The predicted correlation coefficient. The output
            array will have the same shape as the inputs.

        """
        # Verify the validity of input arguments
        if np.any(h < 0):
            raise ValueError("The separation distance must be positive")
        if np.shape(ix1) != np.shape(ix2) or np.shape(ix1) != np.shape(h):
            raise ValueError("The input arguments must all have the same dimensions")

        p1 = self.periods[ix1]
        p2 = self.periods[ix2]

        rho = p1 / p2
        invix = rho > 1.0
        rho[invix] = 1.0 / rho[invix]
        ad = np.abs(h)
        rho = rho * np.exp(-ad / 10)
        h[:] = rho[:]
        return h



#!/usr/bin/env python

import os
import pickle

import numpy as np

import openquake.hazardlib.imt as imt
from openquake.hazardlib.const import StdDev

from esi_shakelib.gmpe.nga_east import NGAEast
from esi_shakelib.multigmpe import stuff_context
from esi_shakelib.utils.contexts import RuptureContext, DistancesContext, SitesContext

home_dir = os.path.dirname(os.path.abspath(__file__))
data_dir = os.path.join(home_dir, "nga_east_data")

stddev_types = [StdDev.TOTAL]
gmpe = NGAEast()

dx = DistancesContext()
dx.rrup = np.logspace(-1, np.log10(2000), 100)

rx = RuptureContext()
sx = SitesContext()

IMTS = [imt.PGA(), imt.PGV(), imt.SA(0.3), imt.SA(1.0), imt.SA(3.0)]

MAGS = [3, 5, 6, 7]

VS30 = [180, 380, 760, 2000]


def update_results():
    # To build the data for testing
    result = {}
    for i in IMTS:
        ikey = i.__str__()
        result[ikey] = {}
        for mag in MAGS:
            rx.mag = mag
            result[ikey][str(mag)] = {}
            for vs30 in VS30:
                sx.vs30 = np.full_like(dx.rrup, vs30)
                sx.sids = np.array(list(range(len(sx.vs30))))
                result[ikey][str(mag)][str(vs30)] = {}
                ctx = stuff_context(sx, rx, dx)
                lmean, lsd = gmpe.get_mean_and_stddevs(ctx, ctx, ctx, i, stddev_types)
                result[ikey][str(mag)][str(vs30)]["lmean"] = lmean.tolist()
                result[ikey][str(mag)][str(vs30)]["lsd"] = lsd[0].tolist()
    # Save results
    pkl_file = os.path.join(data_dir, "nga_east_data.pkl")
    fh = open(pkl_file, "wb")
    pickle.dump(result, fh)
    fh.close()


def test_nga_east():
    # Load test data
    pkl_file = os.path.join(data_dir, "nga_east_data.pkl")
    fh = open(pkl_file, "rb")
    target = pickle.load(fh)
    fh.close()
    for i in IMTS:
        ikey = i.__str__()
        for mag in MAGS:
            rx.mag = mag
            for vs30 in VS30:
                sx.vs30 = np.full_like(dx.rrup, vs30)
                sx.sids = np.array(list(range(len(sx.vs30))))
                ctx = stuff_context(sx, rx, dx)
                lmean, lsd = gmpe.get_mean_and_stddevs(ctx, ctx, ctx, i, stddev_types)
                tmean = np.array(target[ikey][str(mag)][str(vs30)]["lmean"])
                np.testing.assert_allclose(lmean[0], tmean, rtol=1e-6, atol=1e-6)
                tsd = np.array(target[ikey][str(mag)][str(vs30)]["lsd"])
                np.testing.assert_allclose(lsd[0], tsd, rtol=1e-6, atol=1e-6)


if __name__ == "__main__":
    test_nga_east()



# third party imports
import numpy as np
from scipy.ndimage import median_filter
from shapely.geometry import MultiLineString, mapping
from skimage import measure
from openquake.hazardlib import imt

from esi_utils_colors.cpalette import ColorPalette


def contour(imtdict, imtype, filter_size, gmice):
    """
    Generate contours of a specific IMT and return as a Shapely
    MultiLineString object.

    Args:
        container (ShakeMapOutputContainer): ShakeMapOutputContainer
            with ShakeMap output data.
        imtype (str): String containing the name of an Intensity
            Measure Type found in container.
        filter_size (int): Integer filter (see
            https://docs.scipy.org/doc/scipy-0.16.1/reference/generated/scipy.ndimage.filters.median_filter.html)
    Returns:
        list: List of dictionaries containing two fields

                - geometry: GeoJSON-like representation of one of the objects
                  in https://toblerity.org/fiona/manual.html#geometry-types
                - properties: Dictionary of properties describing that
                  feature.

    Raises:
        NotImplementedError -- if the user attempts to contour a data file
            with sets of points rather than grids.
    """  # noqa
    oqimt = imt.from_string(imtype)

    intensity_colormap = ColorPalette.fromPreset("mmi")
    grid = imtdict["mean"]
    metadata = imtdict["mean_metadata"]
    if imtype == "MMI":
        sgrid = grid
        units = "mmi"
    elif imtype == "PGV":
        sgrid = np.exp(grid)
        units = "cms"
    else:
        sgrid = np.exp(grid) * 100.0
        units = "pctg"
    if filter_size > 0:
        fgrid = median_filter(sgrid, size=int(filter_size))
    else:
        fgrid = sgrid

    interval_type = "log"
    if imtype == "MMI":
        interval_type = "linear"

    if np.all(np.isnan(fgrid)):  # data is totally empty; no contours
        intervals = np.array([])
    else:
        grid_min = np.nanmin(fgrid)
        grid_max = np.nanmax(fgrid)
        if grid_max - grid_min:
            intervals = getContourLevels(grid_min, grid_max, itype=interval_type)
        else:  # data is totally flat; don't draw any contours
            intervals = np.array([])

    lonstart = metadata["xmin"]
    latstart = metadata["ymin"]

    lonend = metadata["xmax"]
    if lonend < lonstart:
        lonstart -= 360

    lonspan = np.abs(lonend - lonstart)
    latspan = np.abs(metadata["ymax"] - latstart)
    nlon = metadata["nx"]
    nlat = metadata["ny"]

    line_strings = []  # dictionary of MultiLineStrings and props

    for cval in intervals:
        contours = measure.find_contours(fgrid, cval)
        #
        # Convert coords to geographic coordinates; the coordinates
        # are returned in row, column order (i.e., (y, x))
        #
        new_contours = []
        plot_contours = []
        for ic, coords in enumerate(contours):  # coords is a line segment
            #
            # This greatly reduces the number of points in the contours
            # without changing their shape too much
            #
            coords = measure.approximate_polygon(coords, filter_size / 20)

            mylons = np.around(coords[:, 1] * lonspan / nlon + lonstart, decimals=6)
            mylats = np.around(
                (nlat - coords[:, 0]) * latspan / nlat + latstart, decimals=6
            )

            contours[ic] = np.hstack(
                (mylons[:].reshape((-1, 1)), mylats[:].reshape((-1, 1)))
            )
            plot_contours.append(contours[ic])
            new_contours.append(contours[ic].tolist())

        if len(new_contours):
            mls = MultiLineString(new_contours)
            props = {"value": cval, "units": units}
            if imtype == "MMI":
                pass
            elif imtype == "PGV":
                lcval = np.log(cval)
            else:
                lcval = np.log(cval / 100)
            if gmice:
                mmival = gmice.getMIfromGM(np.array([lcval]), oqimt)[0][0]
            elif imtype == "MMI":
                mmival = cval
            else:
                mmival = 1
            color_array = np.array(intensity_colormap.getDataColor(mmival))
            color_rgb = np.array(color_array[0:3] * 255, dtype=int).tolist()
            props["color"] = "#%02x%02x%02x" % tuple(color_rgb)
            if imtype == "MMI":
                if (cval * 2) % 2 == 1:
                    props["weight"] = 4
                else:
                    props["weight"] = 2
            else:
                props["weight"] = 4
            line_strings.append({"geometry": mapping(mls), "properties": props})
    return line_strings


def getContourLevels(dmin, dmax, itype="log"):
    """
    Get contour levels given min/max values and desired IMT.

    Use itype='log' for any IMT that is logarithmically distributed, such as
    PGA, PGV, and Sa. Linear for MMI.

    Args:
        dmin (float): Minimum value of data to contour.
        dmax (float): Maximum value of data to contour.
        itype (str): Interval type; default is 'log', anythign else
            indicates linear intervals.

    Returns:
        ndarray: Numpy array of contour levels.

    """
    if itype == "log":
        # Within-decade label values
        dec_inc = np.array([1, 2, 5], dtype=float)

        # Upper and lower decades
        lower_dec = np.floor(np.log10(dmin))
        upper_dec = np.ceil(np.log10(dmax))
        # Don't make a crazy number of contours if there are very small
        # values in the input
        if lower_dec < upper_dec - 4:
            lower_dec = upper_dec - 4

        # Array of decades
        decades = np.arange(lower_dec, upper_dec + 1)

        # Construct levels
        levels = np.concatenate([np.power(10, d) * dec_inc for d in decades])
        levels = levels[(levels < dmax) & (levels > dmin)]
        if np.size(levels) == 0:
            levels = np.array([(dmin + dmax) / 2])
    else:
        # MMI contours are every 0.5 units
        levels = np.arange(np.ceil(dmin * 2) / 2, np.floor(dmax * 2) / 2 + 0.5, 0.5)
    return levels



import numpy as np

from openquake.hazardlib import const
from openquake.hazardlib.gsim.base import GMPE, CoeffsTable, registry


class NSHM_AK_Bias_Correction(GMPE):
    """
    Implements the NSHM Alaska bias correction for interface events.
    """

    REQUIRES_SITES_PARAMETERS = set()
    REQUIRES_DISTANCES = set()
    REQUIRES_RUPTURE_PARAMETERS = set()
    DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = ""
    DEFINED_FOR_INTENSITY_MEASURE_TYPES = set()
    DEFINED_FOR_STANDARD_DEVIATION_TYPES = {const.StdDev.TOTAL}
    DEFINED_FOR_TECTONIC_REGION_TYPE = ""
    DEFINED_FOR_REFERENCE_VELOCITY = None

    def __init__(self, gmpe_name, **kwargs):
        super().__init__(**kwargs)
        self.gmpe = registry[gmpe_name]()
        self.COEFFS = self.gmpe.COEFFS
        self.set_parameters()

    def compute(self, ctx: np.recarray, imts, mean, sig, tau, phi):
        self.gmpe.compute(ctx, imts, mean, sig, tau, phi)
        for m, imt in enumerate(imts):
            mean[m] += BIAS_COEFFS[imt]["bias_ak"]


BIAS_COEFFS = CoeffsTable(
    sa_damping=5,
    table="""
    IMT       bias_ak
    PGA     -0.26167
    PGV      0.00833
    0.01    -0.30351
    0.02    -0.35893
    0.03    -0.40397
    0.05    -0.47992
    0.075   -0.54842
    0.1     -0.58624
    0.15    -0.57579
    0.2     -0.53038
    0.25    -0.46174
    0.3     -0.38110
    0.4     -0.23987
    0.5     -0.12297
    0.75     0.04782
    1.0      0.15648
    1.5      0.21896
    2.0      0.22739
    3.0      0.16500
    4.0      0.11244
    5.0      0.05667
    7.5     -0.07493
    10.0    -0.22860
    """,
)
